#####################################################################
# ccat.mpl
#####################################################################
# Description
# -----------
# Implements numeric convex analysis tools.
#
# While this code is robust, much work remains to be done in
# making it efficient.  Additionally, it would be worthwhile to
# implement generalized n dimensional version (but specialed
# 1, 2 and 3 dimension versions).  It would also be interesting
# to implement a numeric subgradient routine (where subgradients
# would be approximated by the convex hull of a finite set of
# points).
#
#####################################################################
# Author		Date		Comment
# ------		----		-------
# CHH			26Nov04		Created.
# CHH			15Feb05		Cleaned up.  Code is conceptually simple
#							however, not as fast as it could be.
#							'In-place' implementations of these
#							transforms would provide a constant
#							factor speedup.  Really, these should
#							be implemented directly in C, and would
#							likely see an order of magnitude speed
#							up!
#
#####################################################################


CCAT := module()

	options `Copyright (c) 2005 by Chris Hamilton.  All rights reserved.`,
		unload=ModuleShutdown:


	export
		# Wrappers to internal functions of the same purpose
		CreateN,
		GetDualRangeN,
		ConjN,
		PlotN,
		ConvexN,
		EvalN,
		ConvexHullN:
	
	local
		# Module Procedures
		ModuleStartup, ModuleShutdown,

		# Package globals
		DefaultPointsPerDimension,

		# Type Definitions
		`_type/NUMERIC`,
		`_type/NPWF`,

		# Pretty printing
		`_print/NPWF`,

		# Array Utilities
		ArrayDim, ArraySize, ArrayMin, ArrayMax,

		# Other Utilities
		MinCoveringRange,
		ConvertRangesToArrays,

		# 1D Stuff
		ConvexHull1d,
		ComputeSlope,
		ComputeSlopeArray1d,
		ComputeConjugateOfConvexHull1d,
		ComputeConjugateOfArray1d,
		IsConvexHull1d,
		CreateLatticeArray1d,
		EvaluateFunction1d,
		GetDualRange1d,
		PlotArray1d,
		
		# 2D Stuff
		IsConvexHull2d,
		ComputeConjugateOfArray2d,
		EvaluateFunction2d,
		GetDualRange2d,
		PlotArray2d:


	#################################################################
	# MODULE PROCEDURES                                             #
	#################################################################


	# Perform module startup
	ModuleStartup := proc()
		global
			NUMERIC, `type/NUMERIC`,
			NPWF, `type/NPWF`, `print/NPWF`:

		unprotect(NUMERIC):
		unprotect(`type/NUMERIC`):
		NUMERIC := 'NUMERIC':
		`type/NUMERIC` := eval(`_type/NUMERIC`):
		protect(NUMERIC):
		protect(`type/NUMERIC`):

		unprotect(NPWF):
		unprotect(`type/NPWF`):
		unprotect(`print/NPWF`):
		NPWF := 'NPWF':
		`type/NPWF` := eval(`_type/NPWF`):
		`print/NPWF` := eval(`_print/NPWF`):
		protect(NPWF):
		protect(`type/NPWF`):
		protect(`print/NPWF`):
	end proc:


	# Perform module shutdown
	ModuleShutdown := proc()
	end proc:


	#################################################################
	# PACKAGE GLOBALS                                               #
	#################################################################


	# Specifies the default number of points to use when
	# creating domains.
	DefaultPointsPerDimension := 25:


	#################################################################
	# TYPE DEFINITIONS                                              #
	#################################################################


	# Defines a numeric, as far as we are concerned for
	# convex functions.  Thus, anything that can be resolved to
	# a floating point number in (-infinity,infinity].
	`_type/NUMERIC` := Or(And(constant,finite),infinity,Float(infinity)):


	# Defines an NPWF.  Essentially, a N dimensional lattice
	# of points and function values.  The actual function format
	# is NPWF(list(array(1,float)),array)
	`_type/NPWF` := proc(
		npwf,
		d::posint
		)
		# Do a very superficial check to ensure it's
		# got the right format.  We don't put much effort
		# here, because it would be done all over the place,
		# and these objects are intended to be generated only
		# by the software itself.  Malformed NPWFs will just
		# cause errors when trying to use them.
		if not evalb(op(0,npwf)='NPWF') or
			not type(op(1,npwf),list) or
			not type(op(2,npwf),'array') then
			return false:
		end if:

		# Dimension specified?
		if nargs > 1 then
			# Ensure XX and F have matching dimension
			if d <> nops(op(1,npwf)) or
				d <> nops([op(2,eval(op(2,npwf)))]) then
				return false:
			end if:
		end if:

		return true:
	end proc:


	#################################################################
	# PRETTY PRINTING                                               #
	#################################################################


	# Pretty print an NPWF function
	`_print/NPWF` := proc( XX, F )
		local i, n, p, x, x1, x2:

		p := NPWF:
		for i from 1 to nops(XX) do
			n := op([2,2],eval(XX[i])):
			x1 := eval(XX[i][1]):
			x2 := eval(XX[i][n]):
			p := p * ``(x[i]=x1..x2):
		end do:

		return p:
	end proc:


	#################################################################
	# ARRAY UTILITIES                                               #
	#################################################################


	# Returns the dimension of the array
	ArrayDim := proc( a::array )
		return nops([op(2,eval(a))]):
	end proc:


	# Returns the number of entries of the array (along all dimensions
	# if no dimension specified)
	ArraySize := proc( a::array, d::posint )
		local s, i:

		# Dimensions provided?
		if nargs > 1 then
			# Invalid dimension?
			if d > ArrayDim(a) then
				error "invalid dimension":
			end if:

			# Get size along given dimension
			s := [op(2,eval(a))]:
			s := s[d]:
			s := eval(op(2,s)-op(1,s)+1):
		else
			# Get entire size
			s := [op(2,eval(a))]:
			s := map(i->eval(op(2,i)-op(1,i)+1),s):
			s := convert(s,`*`):
		end if:

		# Return the size!
		return s:
	end proc:


	# Return the minimum index of the array (along all dimensions if
	# no dimension specified)
	ArrayMin := proc( a::array, d::posint )
		local m, i:

		# Dimensions provided?
		if nargs > 1 then
			# Invalid dimension?
			if d > ArrayDim(a) then
				error "invalid dimension":
			end if:

			# Get min along given dimension
			m := [op(2,eval(a))]:
			m := m[d]:
			m := op(1,m):
		else
			# Get entire min
			m := [op(2,eval(a))]:
			m := map(i->op(1,i),m):
			m := op(m):
		end if:

		# Return the min!
		return m:
	end proc:


	# Return the maximum index of the array (along all dimensions if
	# no dimension specified)
	ArrayMax := proc( a::array, d::posint )
		local m, i:

		# Dimensions provided?
		if nargs > 1 then
			# Invalid dimension?
			if d > ArrayDim(a) then
				error "invalid dimension":
			end if:

			# Get max along given dimension
			m := [op(2,eval(a))]:
			m := m[d]:
			m := op(2,m):
		else
			# Get entire max
			m := [op(2,eval(a))]:
			m := map(i->op(2,i),m):
			m := op(m):
		end if:

		# Return the max!
		return m:
	end proc:


	#################################################################
	# UTILITY CODE                                                  #
	#################################################################


	# Computes the smallest range on the real line covering
	# the given list of ranges.
	MinCoveringRange := proc()
		local xl, xr, i, X:

		# Need at least 1 argument
		if nargs < 1 then
			error "expect at least 1 argument":
		end if:

		# Validate the input data types
		for i from 1 to nargs do
			if not type(args[i],`..`(extended_numeric)) then
				error "expect range(extended_numeric), received %1", args[i]:
			end if:
		end do:

		# Extract all of the end-points
		X := map(i->op(i),[args]):
		xl := min(op(X)):
		xr := max(op(X)):

		# Return the minimum covering range
		return xl..xr:
	end proc:


	# Given an array of ranges, lists, or arrays, will convert these
	# all to arrays.  Newly made arrays will use the given number
	# of points.
	ConvertRangesToArrays := proc(
		RX::list(Or('array'(1,finite),list(finite),`..`(extended_numeric))),
		n::posint
		)
		local XX, i, d:

		# Validate number of points
		if n < 2 then
			error "must provide at least 2 points per axis":
		end if:

		# Get the dimension
		d := nops(RX):

		# Create empty list of arrays
		XX := [0$d]:

		# Ensure each variable range is valid
		for i from 1 to d do
			# Range provided?
			if type(RX[i],`..`) then
				XX[i] := CreateLatticeArray1d(RX[i],n):
			# List or array provided?
			else
				# Grab unique entries, sort, and
				# convert to array
				XX[i] := convert(RX[i],set):
				XX[i] := convert(XX[i],list):
				XX[i] := sort(XX[i]):
				XX[i] := convert(XX[i],array):

				# Ensure at least 2 long!
				if ArraySize(XX[i]) < 2 then
					error "must provide at least 2 points per axis":
				end if:
			end if:
		end do:

		# Return the list of arrays
		return eval(XX):
	end proc:


	#################################################################
	# 1D STUFF                                                      #
	#################################################################


	# Calculates the convex hull of a sorted set of
	# points.
	ConvexHull1d := proc(
		X::'array'(1,finite),
		Y::'array'(1,NUMERIC)
		)
		local i, j,
			CX, CY, mmin, jmin, m, y, dx, dy, i1, i2:

		# Ensure number of points match
		if ArraySize(X) <> ArraySize(Y) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X) <> 1 or ArrayMin(Y) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Skip past any infinities at the beginning
		i1 := 1:
		while i1 <= ArrayMax(X) and evalf(Y[i1]) = Float(infinity) do
			i1 := i1 + 1:
		end do:

		# Skip past any infinities at the end
		i2 := ArrayMax(X):
		while i2 >= i1 and evalf(Y[i2]) = Float(infinity) do
			i2 := i2 - 1:
		end do:

		# Not enough points to bother with
		# full wrapping algorithm?
		j := i2 - i1 + 1:
		if j <= 2 then
			CX := array(1..j):
			CY := array(1..j):
			j := 1:
			for i from i1 to i2 do
				CX[j] := X[i]:
				CY[j] := Y[i]:
				j := j + 1:
			end do:
		else
			# "Beneath-Beyond Algorithm" -> O(|X|)
			# Assumes X is sorted in increasing order.
			# Preparata and Shamos, Computational Geometry, Springer Verlag, 1990
			j := 2:
			(CX[1],CY[1]) := (X[i1],Y[i1]):
			(CX[2],CY[2]) := (X[i1+1],Y[i1+1]):
			for i from i1+2 to i2 do
				# Calculate the projection

				# Both infinity?
				if CY[j] = infinity and CY[j-1] = infinity then
					y := infinity:
				else
					# Calculate the y-value of our hull projected to x=CX[i]
					y := evalf( ((CY[j]-CY[j-1])/(CX[j]-CX[j-1]))*
						(X[i]-CX[j]) + CY[j] ):
				end if:

				# Backup while the new point is beneath
				# our current projected point
				while j > 2 and Y[i] <= y do
					unassign( 'CX[j]' ):
					unassign( 'CY[j]' ):
					j := j-1:
					y := evalf( ((CY[j]-CY[j-1])/(CX[j]-CX[j-1]))*
						(X[i]-CX[j]) + CY[j] ):
				end do:

				# New point is above projection?
				if Y[i] > y then
					# Tack on new point
					j := j+1:
				end if:

				# Store the current point
				(CX[j],CY[j]) := (X[i],Y[i]):
			end do:
		end if:

		# Convert from tables to arrays
		CX := convert(CX,array):
		CY := convert(CY,array):

		# Return the calculated convex hull
		return eval(CX),eval(CY):
	end proc:


	# Returns the slope between two points in R^2
	ComputeSlope := proc(
		x1::finite,
		y1::NUMERIC,
		x2::finite,
		y2::NUMERIC
		)
		local dx, dy:

		# Calculate difference values
		dx := evalf(x2 - x1):
		dy := evalf(y2 - y1):

		# Division by zero?
		if dx = 0 then
			# Negative?
			if dy < 0 then
				return -infinity:
			# Positive?
			else
				return infinity:
			end if:
		else
			return evalf(dy/dx):
		end if:
	end proc:


	# Computes an array of slopes S given an array of
	# points and function evaluations at those points.
	ComputeSlopeArray1d := proc(
		X::'array'(1,finite),
		Y::'array'(1,NUMERIC)
		)
		local i, S:

		# Ensure number of points match
		if ArraySize(X) <> ArraySize(Y) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X) <> 1 or ArrayMin(Y) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# SPECIAL CASE: Less than 2 points?
		if ArraySize(X) < 2 then
			# Return an empty array
			return array(1..0):
		end if:

		# Walk through pairs of points
		for i from 1 to ArraySize(X)-1 do
			S[i] := ComputeSlope(X[i],Y[i],X[i+1],Y[i+1]):
		end do:

		# Conver to array
		S := convert(S,array):

		# Return the array of slopes
		return eval(S):
	end proc:


	# Computes the 1D Fenchel conjugate at the given points given an
	# array of points and function evaluations.  Assumes that (X,Y) form
	# a convex 1D hull, and that X is sorted in ascending order.
	# Returns a single array that is the array of conjugate points.
	ComputeConjugateOfConvexHull1d := proc(
		X::'array'(1,finite),
		Y::'array'(1,finite),
		Xd::'array'(1,finite)
		)
		local S, Yd, i, j:

		# Ensure number of points match
		if ArraySize(X) <> ArraySize(Y) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X) <> 1 or ArrayMin(Y) <> 1 or
			ArrayMin(Xd) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# SPECIAL CASE: If the convex hull is empty,
		# then the function was everywhere +infinity.
		# Thus, the conjugate is everywhere -infinity.
		if ArraySize(X) = 0 then
			Yd := array(1..ArraySize(Xd)):
			for i from 1 to ArraySize(Yd) do
				Yd[i] := -infinity:
			end do:

			return eval(Yd):
		end if:

		# SPECIAL CASE: Singleton convex hull.
		# We have to treat this as a special case
		# because the following code assumes a non-empty
		# array of slopes.
		if ArraySize(X) = 1 then
			Yd := array(1..ArraySize(Xd)):
			for i from 1 to ArraySize(Yd) do
				Yd[i] := X[1]*Xd[1] - Y[1]:
			end do:

			return eval(Yd):
		end if:

		# Calculate the array of slopes
		S := ComputeSlopeArray1d(X,Y):

		# Initialize our loop counters
		(i,j) := (1,1):

		# While Xd[i] < all slopes
		# Thus Xd[i] in grad f(X[1])
		while i <= ArraySize(Xd) and Xd[i] < S[1] do
			Yd[i] := evalf( X[1]*Xd[i] - Y[1] ):
			i := i + 1:
		end do:

		# Walk as far as we can while Xd[i] < S[j]
		while i <= ArraySize(Xd) and j <= ArraySize(S) do

			# Advance through slope array until
			# Xd[i] <= S[j]
			while j <= ArraySize(S) and Xd[i] > S[j] do
				j := j + 1:
			end do:

			# Bail out if we exhausted the slope array
			if j > ArraySize(S) then
				break:
			end if:

			Yd[i] := evalf( X[j]*Xd[i] - Y[j] ):

			i := i + 1:
		end do:

		# At the end of the above loop, if we haven't exhausted
		# Xd then we have exhausted S, thus Xd[i] > all slopes
		j := ArraySize(X):
		while i <= ArraySize(Xd) do
			Yd[i] := evalf( X[j]*Xd[i] - Y[j] ):
			i := i + 1:
		end do:

		# Convert the conjugate points to an array
		Yd := convert(Yd,array):

		# Return the computed conjugate
		return eval(Yd):
	end proc:


	# Conjugates a given numeric function
	ComputeConjugateOfArray1d := proc(
		X::'array'(1,finite),
		Y::'array'(1,NUMERIC),
		Xd::'array'(1,finite)
		)
		local Xc, Yc, Yd:

		# Ensure number of points match
		if ArraySize(X) <> ArraySize(Y) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X) <> 1 or ArrayMin(Y) <> 1 or
			ArrayMin(Xd) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Compute the convex hull
		(Xc,Yc) := ConvexHull1d(X,Y):

		# Compute the conjugate
		Yd := ComputeConjugateOfConvexHull1d(Xc,Yc,Xd):

		# Return the conjugate
		return eval(Yd):
	end proc:


	# Returns TRUE if the given arrays define a 1D convex hull, FALSE
	# otherwise
	IsConvexHull1d := proc(
		X::'array'(1,finite),
		Y::'array'(1,NUMERIC)
		)
		local S, i, d, i1, i2:

		# Ensure number of points match
		if ArraySize(X) <> ArraySize(Y) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X) <> 1 or ArrayMin(Y) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Compute the slope array
		S := ComputeSlopeArray1d(X,Y):

		# NOTE
		# Get lower precision to use for
		# comparison; avoids rounding issues.
		# (eliminates false negatives, but creates
		# some false positives)
		d := max(1,round(Digits/2),Digits-5):

		# Skip undefineds at beginning
		i1 := 1:
		while type(S[i1],undefined) and i1 < ArrayMax(S) do
			i1 := i1 + 1:
		end do:

		# Skip undefineds at end
		i2 := ArrayMax(S):
		while type(S[i2],undefined) and i2 > i1 do
			i2 := i2 - 1:
		end do:

		# Ensure it is non-decreasing
		for i from i1+1 to i2 do
			# If we run into an undefined, then there's
			# something wrong, and it's obviously not
			# convex.
			if type(S[i],undefined) then
				return false:
			end if:

			# Compare with lower digit accuracy
			if evalf(S[i],d) < evalf(S[i-1],d) then
				return false:
			end if:
		end do:

		# If we get here, it's convex
		return true:
	end proc:


	# Gets an 1D array that is a lattice over the given range in R
	CreateLatticeArray1d := proc(
		rx::`..`(extended_numeric),
		n::posint
		)
		local x1, x2, X, i, INF:

		# Get the range end-points
		(x1,x2) := op(rx):

		# Ensure increasing order
		if x1 > x2 then
			(x1,x2) := (x2,x1):
		end if:

		# We'll treat (n-1) as 'infinity' for
		# our purposes.
		INF := n-1:

		# SPECIAL CASE: Handle infinities
		if type(x1,infinity) then
			if type(x2,infinity) then
				(x1,x2) := (evalf(-INF/2),evalf(INF/2)):
			else
				(x1,x2) := (evalf(x2-INF),x2):
			end if:
		elif type(x2,infinity) then
			(x1,x2) := (x1,evalf(x1+INF)):
		end if:

		# SPECIAL CASE: Singleton range.
		# Create a range centered on the singleton, much
		# like Maple does with plots, etc
		if x1 = x2 then
			(x1,x2) := (evalf(x1-1),evalf(x1+1)):
		end if:

		# Create the array
		X := array(1..n):

		# Walk through given number of points
		for i from 1 to n do
			X[i] := evalf(x1 + (i-1)*(x2-x1)/(n-1)):
		end do:

		# Return the lattice
		return eval(X):
	end proc:


	# Evaluates a given function at an array of points
	EvaluateFunction1d := proc(
		f::procedure,
		X::'array'(1,finite)
		)
		local Y, i:

		# Evaluate the function at those points
		Y := map(i->evalf(Re(f(i))),X):

		# Convert to array
		Y := convert(Y,array):

		# Confirm our output is valid!
		if not type(Y,'array'(1,NUMERIC)) then
			error "f must evaluate to type NUMERIC!":
		end if:

		# Return the function evaluated at the given points
		return eval(Y):
	end proc:


	# Returns the largest dual range of practical
	# interest given the primal function.
	GetDualRange1d := proc(
		X::'array'(1,finite),
		Y::'array'(1,NUMERIC)
		)
		local Xc, Yc, xd1, xd2, i:

		# Ensure number of points match
		if ArraySize(X) <> ArraySize(Y) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X) <> 1 or ArrayMin(Y) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Get the convex hull
		(Xc,Yc) := ConvexHull1d(X,Y):

		# SPECIAL CASE: Convex hull is empty or singleton.
		if ArraySize(Xc) < 2 then
			return -infinity..infinity:
		end if:

		# Get the minimum
		xd1 := ComputeSlope(Xc[1],Yc[1],Xc[2],Yc[2]):

		# Get the maximum
		i := ArraySize(Xc):
		xd2 := ComputeSlope(Xc[i-1],Yc[i-1],Xc[i],Yc[i]):

		return xd1..xd2:
	end proc:


	# Plots a pair of arrays
	PlotArray1d := proc(
		X::'array'(1,finite),
		Y::'array'(1,NUMERIC)
		)
		local P, i:

		# Ensure number of points match
		if ArraySize(X) <> ArraySize(Y) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X) <> 1 or ArrayMin(Y) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Create pairs
		for i from 1 to ArraySize(X) do
			P[i] := [X[i],Y[i]]:
		end do:

		# Convert to array
		P := convert(P,array):

		# Plot it!
		return plot(eval(P),args[3..nargs]):
	end proc:


	#################################################################
	# 2D STUFF                                                      #
	#################################################################


	# Determines whether or not a given 2D array is convex
	IsConvexHull2d := proc(
		X1::'array'(1,finite),
		X2::'array'(1,finite),
		F::'array'(2,NUMERIC)
		)
		local i, j, Y:

		# Check dimensions
		if ArraySize(X1) <> ArraySize(F,1) or
			ArraySize(X2) <> ArraySize(F,2) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X1) <> 1 or ArrayMin(X2) <> 1 or
			ArrayMin(F,1) <> 1 or ArrayMin(F,2) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Check convexity along X2 axis
		Y := array(1..ArraySize(X2)):
		for i from 1 to ArraySize(X1) do
			# Extract the 1D subarray
			for j from 1 to ArraySize(X2) do
				Y[j] := F[i,j]:
			end do:

			# Check convexity
			if not IsConvexHull1d(X2,Y) then
				return false:
			end if:
		end do:

		# Check convexity along X1 axis
		Y := array(1..ArraySize(X1)):
		for j from 1 to ArraySize(X2) do
			# Extract the 1D subarray
			for i from 1 to ArraySize(X1) do
				Y[i] := F[i,j]:
			end do:

			# Check convexity
			if not IsConvexHull1d(X1,Y) then
				return false:
			end if:
		end do:

		# If we get here, then we are convex!
		return true:
	end proc:


	# Computes the conjugate of the given 2D array
	ComputeConjugateOfArray2d := proc(
		X1::'array'(1,finite),
		X2::'array'(1,finite),
		F::'array'(2,NUMERIC),
		Y1::'array'(1,finite),
		Y2::'array'(1,finite)
		)
		local G1, G2, i, j, Z, Zd:

		# Check dimensions
		if ArraySize(X1) <> ArraySize(F,1) or
			ArraySize(X2) <> ArraySize(F,2) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X1) <> 1 or ArrayMin(X2) <> 1 or
			ArrayMin(F,1) <> 1 or ArrayMin(F,2) <> 1 or
			ArrayMin(Y1) <> 1 or ArrayMin(Y2) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Compute partial conjugate along x2 axis
		G1 := copy(F):
		G2 := array(1..ArraySize(X1),1..ArraySize(Y2)):
		Z := array(1..ArraySize(X2)):
		for i from 1 to ArraySize(X1) do
			# Extract the 1D subarray
			for j from 1 to ArraySize(X2) do
				Z[j] := G1[i,j]:
			end do:

			# Compute the conjugate
			Zd := ComputeConjugateOfArray1d(X2,Z,Y2):

			# Place values back in 1D array
			for j from 1 to ArraySize(Y2) do
				G2[i,j] := -Zd[j]:
			end do:
		end do:

		# Check convexity along X1 axis
		G1 := copy(G2):
		G2 := array(1..ArraySize(Y1),1..ArraySize(Y2)):
		Z := array(1..ArraySize(X1)):
		for j from 1 to ArraySize(Y2) do
			# Extract the 1D subarray
			for i from 1 to ArraySize(X1) do
				Z[i] := G1[i,j]:
			end do:

			# Compute the conjugate
			Zd := ComputeConjugateOfArray1d(X1,Z,Y1):

			# Place values back in 1D array
			for i from 1 to ArraySize(Y1) do
				G2[i,j] := Zd[i]:
			end do:
		end do:

		# If we get here, then we are convex!
		return eval(G2):
	end proc:


	# Evaluates a given function over a grid of 2D points
	EvaluateFunction2d := proc(
		f::procedure,
		X1::'array'(1,finite),
		X2::'array'(1,finite)
		)
		local F, i, j:

		# Create the output array
		F := array(1..ArraySize(X1),1..ArraySize(X2)):

		# Walk through the array
		for i from 1 to ArraySize(X1) do
			for j from 1 to ArraySize(X2) do
				# Evaluate the function
				F[i,j] := evalf(Re(f(X1[i],X2[j]))):

				# Ensure the output type!
				if not type(F[i,j],NUMERIC) then
					error "f must evaluate to type NUMERIC!":
				end if:
			end do:
		end do:

		# Return the function evaluated at the given points
		return eval(F):
	end proc:


	# Returns the maximum useful dual range given
	# the 2d primal
	GetDualRange2d := proc(
		X1::'array'(1,finite),
		X2::'array'(1,finite),
		F::'array'(2,NUMERIC)
		)
		local ry1, ry2, i, j, Y:

		# Check dimensions
		if ArraySize(X1) <> ArraySize(F,1) or
			ArraySize(X2) <> ArraySize(F,2) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X1) <> 1 or ArrayMin(X2) <> 1 or
			ArrayMin(F,1) <> 1 or ArrayMin(F,2) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Get x2 dual range
		Y := array(1..ArraySize(X2)):
		for i from 1 to ArraySize(X1) do
			for j from 1 to ArraySize(X2) do
				Y[j] := F[i,j]:
			end do:

			if i = 1 then
				ry2 := GetDualRange1d( X2, Y ):
			else
				ry2 := MinCoveringRange( ry2, GetDualRange1d( X2, Y ) ):
			end if:
		end do:

		# Get x2 dual range
		Y := array(1..ArraySize(X1)):
		for j from 1 to ArraySize(X2) do
			for i from 1 to ArraySize(X1) do
				Y[i] := F[i,j]:
			end do:

			if j = 1 then
				ry1 := GetDualRange1d( X1, Y ):
			else
				ry1 := MinCoveringRange( ry1, GetDualRange1d( X1, Y ) ):
			end if:
		end do:

		return ry1, ry2:
	end proc:


	# Plots a 2D array
	PlotArray2d := proc(
		X1::'array'(1,finite),
		X2::'array'(1,finite),
		F::'array'(2,NUMERIC)
		)
		local M, i, j:

		# Check dimensions
		if ArraySize(X1) <> ArraySize(F,1) or
			ArraySize(X2) <> ArraySize(F,2) then
			error "array dimension mismatch":
		end if:

		# Ensure indexed from 1
		if ArrayMin(X1) <> 1 or ArrayMin(X2) <> 1 or
			ArrayMin(F,1) <> 1 or ArrayMin(F,2) <> 1 then
			error "expect arrays indexed from 1":
		end if:

		# Convert to MESH format
		M := copy(F):
		for i from 1 to ArraySize(X1) do
			for j from 1 to ArraySize(X2) do
				M[i,j] := [X1[i],X2[j],F[i,j]]:
			end do:
		end do:

		M := convert(M,listlist):
		M := PLOT3D(MESH(M)):

		return plots[display]( M, args[4..nargs] ):
	end proc:


	#################################################################
	# WRAPPERS                                                      #
	#################################################################


	# Creates an NPWF object
	unprotect(CreateN):
	CreateN := proc(
		f::procedure,
		RX::list(Or('array'(1,finite),list(finite),`..`(extended_numeric))),
		_n::posint
		)
		local i, d, F, XX, n:

		# Get the dimension
		d := nops(RX):

		# Need a non-zero dimension
		if d = 0 then
			error "must have non-zero dimension":
		end if:

		# Right now we only support up to 2 dimensions
		if d > 2 then
			error "currently support only up to 2 dimensions":
		end if:
	
		if nargs > 2 then
			n := max(3,_n):
		else
			n := DefaultPointsPerDimension:
		end if:

		# Convert the possibly mixed input to
		# a list of arrays
		XX := ConvertRangesToArrays(RX,n):

		# Now, evaluate the function over the grid
		if d = 1 then
			F := EvaluateFunction1d(f,XX[1]):
		else
			F := EvaluateFunction2d(f,XX[1],XX[2]):
		end if:

		# Return the NPWF
		return NPWF(eval(XX),eval(F)):
	end proc:
	protect(CreateN):


	# Returns an array of ranges, containing the ranges over
	# which the conjugate is interesting.
	unprotec(GetDualRangeN):
	GetDualRangeN := proc(
		npwf::NPWF
		)
		local i, d, XX, F, RY:

		# Get the NPWF parts
		XX,F := op(npwf):

		# Get the dimension
		d := nops(XX):

		# Only support up to 2 dimensions
		if d > 2 then
			error "currently support only up to 2 dimensions":
		end if:

		# Get the dual range(s)
		if d = 1 then
			RY := [GetDualRange1d(XX[1],F)]:
		else
			RY := [GetDualRange2d(XX[1],XX[2],F)]:
		end if:

		# Return the list of ranges
		return RY:
	end proc:
	protect(GetDualRangeN):

	
	# Conjugates an NPWF object
	unprotect(ConjN):
	ConjN := proc(
		npwf::NPWF,
		_RY::Or(name,list(Or('array'(1,finite),list(finite),`..`(extended_numeric),name))),
		_n::posint
		)
		local i, YY, d, n, G, XX, F, RY, RYd:

		# Get the NPWF parts
		XX,F := op(npwf):

		# Get the dimension
		d := nops(XX):

		# Only support up to 2 dimensions
		if d > 2 then
			error "currently support only up to 2 dimensions":
		end if:

		###############################
		# BEGIN PARSING OF DUAL DOMAINS

		# No _RY specified?  Auto generate
		if nargs < 2 then
			RY := [DEFAULT$d]:
		else
			RY := _RY:
		end if:

		# DEFAULT specified?
		if RY = DEFAULT then
			RY := [DEFAULT$d]:
		end if:

		# RY too big?
		if nops(RY) > d then
			error "too many dual ranges specified":
		end if:

		# RY too small?  Pad out with DEFAULT
		if nops(RY) < d then
			RY := [eval(op(RY)),DEFAULT$(d-nops(RY))]:
		end if:

		# Have any DEFAULT parameters to fill in?
		if has(RY,DEFAULT) then
			# Automatically determine the interesting
			# dual ranges
			RYd := GetDualRangeN(npwf):

			# Fill in any of the ones left to
			# automatically determined.
			for i from 1 to nops(RY) do
				if RY[i] = DEFAULT then
					RY[i] := RYd[i]:					
				end if:
			end do:
		end if:

		# At this point, if some part of DY is still malformed
		# from the input, then it'll be caught by the
		# the call to ConvertRangesToArrays.

		# END PARSING OF DUAL DOMAINS
		#############################

		if nargs > 2 then
			n := max(3,_n):
		else
			n := DefaultPointsPerDimension:
		end if:

		# Convert the possibly mixed input to
		# a list of arrays
		YY := ConvertRangesToArrays(RY,n):

		# Compute the conjugate
		if d = 1 then
			G := ComputeConjugateOfArray1d(XX[1],F,YY[1]):
		else
			G := ComputeConjugateOfArray2d(XX[1],XX[2],F,YY[1],YY[2]):
		end if:

		# Return the conjugate NPWF object
		return NPWF(eval(YY),eval(G)):
	end proc:
	protect(ConjN):


	# Plots an NPWF object
	unprotect(PlotN):
	PlotN := proc(
		npwf::NPWF
		)
		local XX, F, d, p, opts:

		# Get the NPWF parts
		XX,F := op(npwf):

		# Get the dimension
		d := nops(XX):

		# Only support up to 2 dimensions
		if d > 2 then
			error "currently support only up to 2 dimensions":
		end if:

		# Get any options to pass through
		opts := args[2..nargs]:

		# Do the plot
		if d = 1 then
			p := PlotArray1d(XX[1],F,opts):
		else
			p := PlotArray2d(XX[1],XX[2],F,opts):
		end if:

		return p:
	end proc:
	protect(PlotN):


	# Determines whether or not the given NPWF represents
	# a convex function
	unprotect(ConvexN):
	ConvexN := proc(
		npwf::NPWF
		)
		local XX,F,d:

		# Get the NPWF parts
		XX,F := op(npwf):

		# Get the dimension
		d := nops(XX):

		# Only support up to 2 dimensions
		if d > 2 then
			error "currently support only up to 2 dimensions":
		end if:

		if d = 1 then
			return IsConvexHull1d( XX[1], F ):
		else
			return IsConvexHull2d( XX[1], XX[2], F ):
		end if:
	end proc:
	protect(ConvexN):


	# Evaluates a given NPWF function at a point. Uses
	# linear interpolation when off the grid.
	# NOTE: This routine will work for arbitrary dimensions.
	unprotect(EvalN):
	EvalN := proc(
		npwf::NPWF,
		X::list(finite)
		)
		local i, j, k, XX, F, d, l, S, fx:

		# Get the NPWF parts
		XX,F := op(npwf):

		# Get the dimension
		d := nops(XX):

		# Ensure the dimensions match
		if d <> nops(X) then
			error "dimension mismatch":
		end if:

		# Create blank list
		S := {[1,[]]}:

		# Walk through each dimension
		for i from 1 to d do
			# Walk through the domain array, to
			# find which indices our point falls between.
			j := 1:
			while j <= ArraySize(XX[i]) and X[i] > XX[i][j] do
				j := j+1:
			end do:		

			# Of the end of the array?
			if j > ArraySize(XX[i]) then
				return infinity:
			else
				# Equality?
				if X[i] = XX[i][j] then
					S := map(k->[k[1],[op(k[2]),j]],S):
				else
					# Of the beginning of the array?
					if j = 1 then
						return infinity:
					else
						# Determine the convex combination weight between
						# XX[i][j-1] and XX[i][j] that gives us
						# X[i].
						l := evalf((X[i]-XX[i][j-1])/(XX[i][j]-XX[i][j-1])):

						# Update out set of points, and maintain
						# the convex combination weights to each
						S := map(k->[(1-l)*k[1],[op(k[2]),j-1]],S) union
							map(k->[l*k[1],[op(k[2]),j]],S):
					end if:
				end if:
			end if:
		end do:
		S := convert(S,list):

		# Now, extract the weighted value of the function at
		# each of the points in S.
		S := map(i->i[1]*F[op(i[2])],S):

		# Perform the convex sum
		fx := convert(S,`+`):		

		return fx:
	end proc:
	protect(EvalN):


	# Computes the convex hull of the given NPWF.
	unprotect(ConvexHullN):
	ConvexHullN := proc(
		npwf::NPWF
		)
		local i, j, F, G, XX, g, X, Y:

		# One dimensional?
		if type(npwf,NPWF(1)) then
			XX,F := op(npwf):
			XX,F := ConvexHull1d(XX[1],F):

			return NPWF([eval(XX)],eval(F))
		end if:

		# Two dimensional?
		if type(npwf,NPWF(2)) then
			XX,F := op(npwf):
			G := copy(F):

			for i from 1 to ArraySize(XX[1]) do
				Y := copy(XX[2]):
				for j from 1 to ArraySize(XX[2]) do
					Y[j] := G[i,j]:
				end do:

				X,Y := ConvexHull1d(XX[2],Y):
				g := NPWF([eval(X)],eval(Y)):

				for j from 1 to ArraySize(XX[2]) do
					G[i,j] := EvalN( g, [XX[2][j]] ):
				end do:
			end do:

			for j from 1 to ArraySize(XX[2]) do
				Y := copy(XX[1]):
				for i from 1 to ArraySize(XX[1]) do
					Y[i] := G[i,j]:
				end do:

				X,Y := ConvexHull1d(XX[1],Y):
				g := NPWF([eval(X)],eval(Y)):

				for i from 1 to ArraySize(XX[1]) do
					G[i,j] := EvalN( g, [XX[1][i]] ):
				end do:
			end do:

			return NPWF(XX,eval(G)):
		end if:
	end proc:
	protect(ConvexHullN):


	#################################################################
	# MODULE BODY                                                   #
	#################################################################


	# Start up this module (export typedefs, pretty printers,
	# simplifiers, etc)
	ModuleStartup():

end module:
