#####################################################################
# scat.mpl
#####################################################################
# Description
# -----------
# Implements a Symbolic Convex Analysis Toolkit, for performing
# common convex analysis operations such as sub-differentiation,
# integration, conjugation, infimal convolution, convexity testing,
# etc.  Is implicitly limited to working with lower-semicontinuous
# functions f : R^n -> R.
#
# Integrates with Maple's pretty-printing, conversion,
# and simplification facilities.  Includes support for common
# arithmetic operations and plotting.
#
# Notes
# -----
# Simplification code is a little weak right now.  Sometimes the
# output won't be a valid PWF or SD.  Additionally, we could try
# harder to make certain PWF simplifications succeed (variable
# reordering, dimension upsampling, etc).
#
#####################################################################
# Author		Date		Comment
# ------		----		-------
# CHH			21Apr04		Created (from earlier non-module code)
# CHH			23Apr04		Added pivoting code for n-dimensional
#							conjugation.
# CHH			24Apr04		Fixed non-continuous open-interval
#							problem after pivoting.
# CHH			25Apr04		Fixed boundary point problem after
#							pivoting.
#							Fixed sometimes selecting incorrect
#							inverse in SelectInverse.
# CHH			27Apr04		Fixed bug in PivotRegion2dMonotone
#							(fm>0 gm=0 case was incorrect)
#							Fixed appearance of main variable in
#							points after pivoting (PivotRegion)
#							Fixed RegionCompare bug.
#							Added PartialConj and Pivot procs.
# CHH			12Dec04		Numerous little bug fixes, and user
#							friendly interface tweaks to exported
#							functions.
#							Added conditional pretty-printing of
#							values at points based on dimension of
#							PWF.
# CHH			28Feb05		Added pwf <-> piecewise/region conversion
#							functionality.
# MKT                   28Nov12         Added 'package' to module options (for Maple 16)
#
#####################################################################


SCAT := module()

	options `Copyright (c) 2005 by Chris Hamilton.  All rights reserved.`,
		package, unload=ModuleShutdown:


	export
		PwfToPiecewise, PwfToRegions, RegionsToPwf,
		Plot,
		Assume, Additionally, Assumptions,
		Eval, Equal, Limit, Subs,
		SubDiff, Integ,
		PartialConj, Pivot, Conj,
		Convex, InfConv:
	
	local
		# Module Procedures
		ModuleStartup, ModuleShutdown,

		# Assumption Facilities
		AssumeResetStack, AssumeStack, AssumeCur,
		AssumePush, AssumePop, AssumeTrim,

		# Basic Typedefs (these get exported in ModuleStartup)
		`_type/EXPR`, `_type/BOOL`, `_type/ASSUMPTION`, `_type/VARLIST`,

		# Conditional Boolean Code
		GetVars, SimplifyRelation, SimpleEvalRel,
		AlwaysTrue, AlwaysFalse, MayBeTrue, MayBeFalse, EvalRel,

		# Utility Code
		SimplifyExpr, AreEqual, IsZero, IsZeroVector, GetPointsInOpenInterval,
		GetPointInOpenInterval, MaxIntegersInRealRange,
		GetIntegersInRealRange, GetParameterizedPointsInOpenInterval,
		GetSolutionsInOpenInterval,
		
		# General Inversion Code
		GetAllInverses, SelectInverse, TestInverse, InvertMonotoneFunction,
		GeneralInverseOnOpenInterval,

		# PWF Code		
		`_type/PWF`, `_print/PWF`, PwfPrint, `PWFPL/Simplify`, `PWF/Simplify`,
		PiecewiseToPiecelist, PiecewiseToPwfPl, `_convert/PWF`,
		PwfBopAdd, PwfBopMult, PwfBopPow,
		PwfApplyBopPlPl, PwfApplyBopPlExpr, PwfApplyBopExprPl, PwfApplyBop,
		PwfApplyUopPl, PwfApplyUop,
		`_simplify/PWF`,

		# SD Code
		`_type/SD`, `_print/SD`, `SD/Simplify`, `_convert/SD`,
		SdBopAdd, SdBopMult, SdBopPow,
		SdFillPoints, SdApplyBop, SdApplyUop,
		`_simplify/SD`,

		# Evaluation Code		
		`PWFPL/Eval`, `PWF/Eval`, `SD/Eval`,

		# Limit Taking Code
		`PWFPL/Limit`, `PWF/Limit`, `SD/Limit`,

		# Plotting Code		
		EvaluateRange, `PWF/Plot1D`, `PWF/Plot2D`, `SD/Plot`,

		# Equality Testing Code		
		`PWFPL/Equal`, `PWF/Equal`, `SD/Equal`,

		# Subdifferentiation Code
		`PWFPL/SubDiff`,

		# Integration Code
		`SD/Integrate`,

		# 1D Conjugation Code
		`PWFPL/Conj1D`,
		
		# Pivoting Code
		RegionDeps, PivotRegion2dMonotone, PivotRegion2d, PivotRegion,
		PwfPlToRegions, RegionCompare, RegionsToPwfPl,
		`PWFPL/PivotFix1D`, `PWFPL/PivotFix`, `PWFPL/FixBoundaryPoints`,
		`PWFPL/Pivot`,
		
		# Conjugation Code
		`PWFPL/ConjRecur`, `PWFPL/Negate`, `PWFPL/Conj`:


	#################################################################
	# MODULE PROCEDURES                                             #
	#################################################################


	# Perform module startup
	ModuleStartup := proc()
		global EXPR, `type/EXPR`,
			BOOL, `type/BOOL`,
			VARLIST, `type/VARLIST`,
			ASSUMPTION, `type/ASSUMPTION`,
			PWF, `type/PWF`, `print/PWF`, `convert/PWF`, `simplify/PWF`,
			SD, `type/SD`, `print/SD`, `convert/SD`, `simplify/SD`:

		# Initialize assumption stuff
		AssumeStack := [{}]:
		AssumeCur := {}:

		# Protect the Maple 'property' data type (used
		# by our assumption facilities)
		protect(property):
		protect(`type/property`):

		unprotect(EXPR):
		unprotect(`type/EXPR`):
		EXPR := 'EXPR':
		`type/EXPR` := eval(`_type/EXPR`):
		protect(EXPR):
		protect(`type/EXPR`):

		unprotect(BOOL):
		unprotect(`type/BOOL`):
		BOOL := 'BOOL':
		`type/BOOL` := eval(`_type/BOOL`):
		protect(BOOL):
		protect(`type/BOOL`):

		unprotect(ASSUMPTION):
		unprotect(`type/ASSUMPTION`):
		ASSUMPTION := 'ASSUMPTION':
		`type/ASSUMPTION` := eval(`_type/ASSUMPTION`):
		protect(ASSUMPTION):
		protect(`type/ASSUMPTION`):

		unprotect(VARLIST):
		unprotect(`type/VARLIST`):
		VARLIST := 'VARLIST':
		`type/VARLIST` := eval(`_type/VARLIST`):
		protect(VARLIST):
		protect(`type/VARLIST`):

		unprotect(PWF):
		unprotect(`type/PWF`):
		unprotect(`print/PWF`):
		unprotect(`convert/PWF`):
		unprotect(`simplify/PWF`):
		PWF := 'PWF':
		`type/PWF` := eval(`_type/PWF`):
		`print/PWF` := eval(`_print/PWF`):
		`convert/PWF` := eval(`_convert/PWF`):
		`simplify/PWF` := eval(`_simplify/PWF`):
		protect(PWF):
		protect(`type/PWF`):
		protect(`print/PWF`):
		protect(`convert/PWF`):
		protect(`simplify/PWF`):

		unprotect(SD):
		unprotect(`type/SD`):
		unprotect(`print/SD`):
		unprotect(`convert/SD`):
		unprotect(`simplify/SD`):
		SD := 'SD':
		`type/SD` := eval(`_type/SD`):
		`print/SD` := eval(`_print/SD`):
		`convert/SD` := eval(`_convert/SD`):
		`simplify/SD` := eval(`_simplify/SD`):
		protect(SD):
		protect(`type/SD`):
		protect(`print/SD`):
		protect(`convert/SD`):
		protect(`simplify/SD`):
	end proc:


	# Perform module shutdown
	ModuleShutdown := proc()
	end proc:


	#################################################################
	# ASSUMPTION FACILITIES                                         #
	#################################################################


	# Resets the assumption stack
	AssumeResetStack := proc()
		AssumeCur := AssumeStack[1]:
		AssumeStack := [AssumeStack[1]]:

		return:
	end proc:


	# Pushes a set of assumptions on to the
	# assumption stack.  Assumes the input is valid.
	AssumePush := proc( as )
		AssumeStack := [op(AssumeStack),as]:
		AssumeCur := AssumeCur union as:

		# HACK: Remove any assumptions involving
		# infinities, which don't really tell us
		# anything additional, and generally confuse
		# the hell out of the assumption code
		# (is, coulditbe, etc)
		AssumeCur := remove(has,AssumeCur,infinity):

		return:
	end proc:


	# Pops the top set of assumptions from the
	# assumption stack
	AssumePop := proc()
		local n, as:

		n := nops(AssumeStack):

		# Can't pop the global assumptions
		if n = 1 then
			return:
		end if:

		# Get the set we're removing
		as := AssumeStack[n]:

		# Remove it from the stack
		AssumeStack := subsop(n=NULL,AssumeStack):

		# If this set isn't empty, remove it from the
		# current assumption set as well.
		if nops(as) > 0 then
			AssumeCur := `union`(op(AssumeStack)):
		end if:

		# HACK: Remove any assumptions involving
		# infinities, which don't really tell us
		# anything additional, and generally confuse
		# the hell out of the assumption code
		# (is, coulditbe, etc)
		AssumeCur := remove(has,AssumeCur,infinity):

		return:
	end proc:


	# Removes unapplicable assumptions from a PWF or SD.
	AssumeTrim := proc( f )
		local pl, vl, al, V, i:

		# Get constituents
		(pl,vl,al) := op(f):

		# Only keep valid assumptions
		al := select(i->evalb(type(i,ASSUMPTION)),al):

		# Remove any assumptions with variables NOT occuring
		# in the object.
		V := GetVars([pl,vl]):
		al := remove(i->evalb(nops(GetVars(i) minus V)>0),al):

		# Remove any assumptions with NO variables
		al := remove(i->evalb(nops(GetVars(i)) = 0),al):

		# Remove any ?::complex assumptions
		al := remove(i->evalb(op(2,i)=complex),al):

		# Remove any assumptions on vl variables,
		# EXCEPT ?::real
		V := convert(vl,set):
		al := remove(i->evalb(nops(GetVars(i) intersect V)=1 and op(2,i)<>real),al):

		return subsop(3=al,f):
	end proc:


	# Sets global assumptions, or assumptions that
	# travel with a PWF or SD.
	# First argument may be a PWF or SD, but otherwise
	# expects all arguments to be of type ASSUMPTION.
	Assume := proc()
		local f, as, ias, i:

		# Is the first argument a PWF or SD?
		if nargs > 0 then
			if type(args[1],Or(PWF,SD)) then
				f := args[1]:
				as := {args[2..nargs]}:
			else
				f := NULL:
				as := {args}:
			end if:
		else
			f := NULL:
			as := {args}:
		end if:

		# Global assumptions?
		if f = NULL then
			AssumeStack[1] := as:
			return AssumeStack[1]:
		# Local PWF assumptions?
		else
			# Create implied assumptions
			ias := GetVars([op(1..2,f)]):
			ias := map(i->i::real,ias):

			# Merge provided and implied assumptions
			as := as union ias:
			f := subsop(3=as,f):

			# Remove any unapplicable and invalid assumptions
			f := AssumeTrim(f):

			return f:
		end if:
	end proc:
	protect(Assume):


	# Adds to the global or local PWF or SD assumptions.
	# First argument may be a PWF or SD, but otherwise
	# expects all arguments to be of type ASSUMPTION.
	Additionally := proc()
		local f, as, i:

		# Is the first argument a PWF or SD?
		if nargs > 0 then
			if type(args[1],Or(PWF,SD)) then
				f := args[1]:
				as := {args[2..nargs]}:
			else
				f := NULL:
				as := {args}:
			end if:
		else
			f := NULL:
			as := {args}:
		end if:

		# Validate the assumptions
		for i from 1 to nops(as) do
			if not type(as[i],ASSUMPTION) then
				error "expect ASSUMPTION, received %1", as[i]:
			end if:
		end do:

		# Global assumptions?
		if f = NULL then
			AssumeStack[1] := AssumeStack[1] union as:
			return AssumeStack[1]:
		# Local PWF assumptions?
		else
			# Merge provided assumptions
			as := as union op(3,f):
			f := subsop(3=as,f):

			# Remove any unapplicable assumptions
			f := AssumeTrim(f):

			return f:
		end if:

		return:
	end proc:
	protect(Additionally):


	# Returns the current global or local PWF assumptions.
	# If there's an argument, it must be a PWF.
	Assumptions := proc( f::Or(PWF,SD) )
		if nargs > 0 then
			return op(3,f):
		else
			return AssumeStack[1]:
		end if:
	end proc:
	protect(Assumptions):


	#################################################################
	# BASIC TYPEDEFS                                                #
	#################################################################


	# Define EXPR data type
	# (like algebraic, but a little more restrictive)
	`_type/EXPR` := proc( e )
		if type(e,function) then
			if op(1,e) = piecewise or op(1,e) = PWF or
				op(1,e) = SD then
				return false:
			end if:
		end if:

		if not type(e,Or(integer,fraction,float,symbol,
			indexed,`+`,`*`,`^`,`**`,function,`!`,name)) then
			return false:
		end if:

		return true:
	end proc:


	# Define BOOL data type
	# (true or false only)
	`_type/BOOL` := x -> evalb((x=true) or (x=false)):


	# Defines an ASSUMPTION
	`_type/ASSUMPTION` := proc( a )
		# Inequalities are valid
		if type(a,`<`) or type(a,`<=`) then
			# Must have at least one name in it!
			if nops(indets(a,name)) = 0 then
				return false:
			end if:
		# `::` objects are valid
		elif type(a,`::`) then
			# Left-hand side must be a name
			if not type(op(1,a),name) then
				return false:
			end if:

			# Must have property for right-hand side
			if not type(op(2,a),property) then
				return false:
			end if:
		# Nothing else is valid!
		else
			return false:
		end if:

		return true:
	end proc:


	# Defines a VARLIST
	`_type/VARLIST` := proc( vl, dim::nonnegint )
		if not type(vl,list(name)) then
			return false:
		end if:

		if nops(convert(vl,set)) <> nops(vl) then
			return false:
		end if:

		if nargs > 1 then
			if dim <> nops(vl) then
				return false:
			end if:
		end if:

		return true:
	end proc:


	#################################################################
	# CONDITIONAL BOOLEAN CODE                                      #
	#################################################################

	# Returns the set of all unbound variables occurring
	# in the given parameters.
	GetVars := proc()
		return indets([args],And(Or(name,symbol),Not(constant))):
	end proc:


	# Simplifies a relation.  Provides short cut evaluations
	# of certain relations enforcing some things which aren't generally
	# true, but are in the context of SCAT.
	SimplifyRelation := proc( _r )
		local r:

		r := _r:

		# We don't like Float(infinity)'s hanging around
		r := subs(Float(infinity)=infinity,r):
		r := subs(Float(-infinity)=-infinity,r):

		# Simplify things as much as we can
		r := simplify(r):

		# If we contain any infinities, isolate them!
		# This puts relations into a form compatible
		# to our following simplifications.
		if has(r,infinity) and op(1,r) <> -infinity and
			op(1,r) <> infinity then
			r := isolate(r,infinity):
		end if:

		# Rules we enforce
		# infinity < anything -> false
		# anything < -infinity -> false
		# Not(infinity) < infinity -> true
		# -infinity < Not(-infinity) -> true
		# anything <= infinity -> true
		# -infinity <= anything -> true
		# !infinity = +/-infinity -> false

		if type(r,`<`) then
			if op(1,r) = infinity then
				r := false:
			elif op(2,r) = -infinity then
				r := false:
			elif op(2,r) = infinity then
				if op(1,r) = infinity then
					r := false:
				else
					r := true:
				end if:
			elif op(1,r) = -infinity then
				if op(2,r) = -infinity then
					r := false:
				else
					r := true:
				end if:
			end if:
		elif type(r,`<=`) then
			if op(2,r) = infinity then
				r := true:
			elif op(1,r) = -infinity then
				r := true:
			end if:
		elif type(r,`=`) then
			if has(op(1,r),infinity) <> has(op(2,r),infinity) then
				return false:
			end if:
		end if:

		# Return the simplified relation
		return r:
	end proc:

	# Evaluates a relation without reference to any
	# assumption facilities.  Ignores Maple results that
	# are based on form of the operands rather than the
	# values of the operands.  If unable to determine
	# anything, return the original expression.  Doesn't
	# do simplification or anything.
	SimpleEvalRel := proc( _r )
		local b, r:

		# Get a copy
		r := _r:

		# Do a simple evaluation
		b := evalb(r):

		# Occasionally, evalf will help
		if not type(b,BOOL) then
			b := evalb(evalf(r)):
		end if:

		# Did we get an evaluated result?
		if type(b,BOOL) then
			# Can't trust 'false' '=' results
			if b = false and op(0,r) = `=` then
				return r:
			end if:

			# Can't trust 'true' '<>' results
			if b = true and op(0,r) = `<>` then
				return r:
			end if:

			# If we get here, we can trust the
			# answer
			return b:
		end if:

		# Maple couldn't figure it out!
		return r:
	end proc:

	# Evaluates whether or not an expression is ALWAYS true: a single
	# counter-example is sufficient to return false.  Throws an
	# error if unable to determine.
	AlwaysTrue := proc( _r )
		local r:

		# Simplify the relation
		r := SimplifyRelation(_r):

		# Done?
		if type(r,BOOL) then
			return r:
		end if:

		# Next, try a simple evaluation
		r := SimpleEvalRel(r):

		# Done?
		if type(r,BOOL) then
			return r:
		end if:

		# Now, try to see if the assumption facilities can
		# answer this question
		r := is(r) assuming op(AssumeCur):

		# Done?
		if type(r,BOOL) then
			return r:
		end if:
		
		# If we get here, then the question is undecidable given
		# the current information
		error "unable to evaluate relation: %1", _r:
	end proc:


	# Evaluates whether or not an expression is ALWAYS false: a single
	# counter-example is sufficient to return false.  Throws an
	# error if unable to determine.
	AlwaysFalse := proc( _r )
		local r:

		# Simplify the relation
		r := SimplifyRelation(_r):

		# Done?
		if type(r,BOOL) then
			return r:
		end if:

		# Next, try a simple evaluation
		r := SimpleEvalRel(r):

		# Done?
		if type(r,BOOL) then
			return not(r):
		end if:

		# Is it possible for the statement to be true ever?
		r := coulditbe(r) assuming op(AssumeCur):

		# Done?
		if type(r,BOOL) then
			return not(r):
		end if:
		
		# If we get here, then the question is undecidable given
		# the current information
		error "unable to evaluate relation: %1", _r:
	end proc:


	# Evaluates whether or not a relation could possibly
	# be true.  Returns true if it is possible, false otherwise.
	MayBeTrue := proc()
		return not(AlwaysFalse(args)):
	end proc:


	# Evaluates whether or not a relation could possibly
	# be false.  Returns true if it is possible, false otherwise.
	MayBeFalse := proc()
		return not(AlwaysTrue(args)):
	end proc:


	# Evaluates a relation, taking into account any
	# assumptions.  Returns true if the statement is ALWAYS
	# true, false if it is ALWAYS false, and returns an error
	# if neither, or unable to determine.
	EvalRel := proc( _r::relation )
		local b, r:

		# Simplify the relation
		r := SimplifyRelation(_r):

		# Done?
		if type(r,BOOL) then
			return r:
		end if:

		# Next, try a simple evaluation
		r := SimpleEvalRel(r):

		# Done?
		if type(r,BOOL) then
			return r:
		end if:

		# Now, try to see if the assumption facilities can
		# answer this question

		# Is this statement always true?
		b := is(r) assuming op(AssumeCur):
		if b = true then
			return true:
		end if:

		# Is it always false?
		b := coulditbe(r) assuming op(AssumeCur):
		if b = false then
			return false:
		end if:
		
		# If we get here, then the question is undecidable given
		# the current information
		error "unable to evaluate relation: %1", _r:
	end proc:


	#################################################################
	# UTILITY CODE                                                  #
	#################################################################


	# Simplifies an expression that is explicitly known
	# to be real-valued.
	# TODO: Anything better that we can do here?
	SimplifyExpr := proc( _e )
		local e, i:

		# Grab a copy of the expression
		e := _e:

		# HACK: Has infinities?  Try a numeric evaluation
		# shortcut.
		if has(e,infinity) then
			i := e:
			i := evalf(i):
			if not type(i,float) then
				i := simplify(i) assuming op(AssumeCur):
			end if:
			if type(i,float) then
				i := Re(i):
				if i = Float(infinity) then
					return infinity:
				elif i = Float(-infinity) then
					return -infinity:
				end if:
			end if:
		end if:

		# HACK: If we have complex trigs, use
		# exponentials instead: MUCH easier to simplify.
		if has(e,I) then
			e := convert(e,exp):
		end if:

		# HACK: If we have radicals, then attempt to rationalize
		if hastype(e,radical) then
			e := rationalize(e):
		end if:

		# HACK: Does this have (-1)^numeric? If so, then
		# we use a bit of hack to coax solutions.
		if hastype(e,(-1)^fraction) then
			i := e:
			try
				i := ln(i):
				i := expand(i):
				i := simplify(i) assuming op(AssumeCur):
				i := exp(i):
				i := expand(i):
				e := i:
			catch :
			end try:
		end if:

		# Simplify!
		e := simplify(e) assuming op(AssumeCur):

		# Get rid of Float(infinity)'s!
		e := subs(Float(infinity)=infinity,Float(-infinity)=-infinity,e):

		return e:
	end proc:


	# Uses a series of tests (including probabilitistic
	# tests) to see if the given statements are equal.
	AreEqual := proc( _e1, _e2 )
		local b, e, e1, e2:

		# Try a blatant compare first
		if _e1 = _e2 then
			return true:
		end if:

		# Grab simplified copies
		e1 := SimplifyExpr(_e1):
		e2 := SimplifyExpr(_e2):
		e := e1 = e2:

		# Apply any SCAT specific simplifications
		e := SimplifyRelation(e):

		# Done?
		if type(e,BOOL) then
			return e:
		end if:

		# Next, try a simple evaluation
		e := SimpleEvalRel(e):

		# Done?
		if type(e,BOOL) then
			return e:
		end if:

		# Try evaluating as a float
		b := abs(evalf(e1-e2,Digits+10)):

		# Did it return a float?
		if type(b,float) then
			# Close to zero?
			b := evalb( b < 10^(-Digits) ):

			# Done?
			if type(b,BOOL) then
				return b:
			end if:
		end if:

		# Try probabilistic testing
		# Can only trust true results.  False results
		# indicate that a counter-example exists, but under
		# our constrained assumptions this counter-example
		# might not be possible.  So we can't trust
		# false results.
		b := testeq(e):

		# Done?
		if b = true then
			return b:
		end if:

		# Try testing with assumptions
		b := EvalRel(e):
		if type(b,BOOL) then
			return b:
		end if:

		# Couldn't figure it out!
		error "unable to determine if %1 = %2", _e1, _e2:
	end proc:


	# Tests to see if a given expression is zero.
	IsZero := proc( e )
		return AreEqual(e,0):
	end proc:


	# Uses probabilistic and numeric tests to see if the
	# given vector is essentially a zero vector.  Returns
	# true, false or throws an error.
	IsZeroVector := proc( _v )
		local v, i:

		# Grab a copy, and ensure it's a set, killing
		# any duplicate entries.
		v := convert(_v,set):

		# Walk through the entries
		for i from 1 to nops(v) do
			# Test each entry.  If any entry
			# returns false, we're done.
			if IsZero(v[i]) = false then
				return false:
			end if:
		end do:

		# If we get here, each entry tested
		# as zero, so return true
		return true:
	end proc:


	# Returns n-points in the given open interval (a,b)
	GetPointsInOpenInterval := proc( a, b, n )
		local X, i, j:

		# Check for infinities at end-points
		if a = -infinity and b = +infinity then
			X := [seq( i-floor(n/2), i=0..n-1 )]:
		elif a = -infinity then
			X := [seq( b-n+i, i=0..n-1 )]:
		elif b = infinity then
			X := [seq( a+i+1, i=0..n-1 )]:
		else
			X := simplify([seq( (a*(n+1-i)+b*i)/(n+1), i=1..n )]):
		end if:

		# Return the set of points in the interval
		return X:
	end proc:


	# Returns 1 point in the given open interval
	GetPointInOpenInterval := proc( a, b )
		local X:
		X := GetPointsInOpenInterval(a,b,1):
		return X[1]:
	end proc:


	# Converts a RealRange(a,b) to an ordered list of
	# integers with the given range.  Returns an error
	# if there are too many (greater than MaxIntegersInRealRange)
	# such integers.
	MaxIntegersInRealRange := 500:
	GetIntegersInRealRange := proc( rr )
		local X, A, B, a, b, i:

		# Check for infinities
		if has(rr,infinity) then
			error "infinitely many solutions: %1", rr:
		end if:

		# Get the operands, removing Open() from them
		# if necessary
		(A,B) := op(rr):
		if op(0,A) = Open then A := op(A): end if:
		if op(0,B) = Open then B := op(B): end if:

		# Get the ceiling and floor of the range
		# endpoints
		a := ceil(A):
		b := floor(B):

		# Ensure they're integers!
		if not(type(a,integer)) then
			error "unable to determine ceil(%1)", A:
		end if:
		if not(type(b,integer)) then
			error "unable to determine floor(%1)", B:
		end if:

		# If the lower bound was an open integer, adjust our
		# endpoint
		if a = A and op([1,0],rr) = Open then
		a := a + 1:
		end if:

		# If the upper bound was an open integer, adjust our
		# endpoint
		if b = B and op([2,0],rr) = Open then
		b := b - 1:
		end if:

		# Too many solutions?
		if b - a + 1 > MaxIntegersInRealRange then
			error "too many integers in range: %1", rr:
		end if:

		# Return all of the integers in the range
		X := {seq(i,i=a..b)}:

		return X:
	end proc:


	# Given an expression p parameterized in the variables V (variables
	# of the form _Zxx, _Bxx, or _NNxx), determines all such points
	# that fall within the open interval (a,b)
	GetParameterizedPointsInOpenInterval := proc( p, V, a, b )
		options remember:
		local X, V2, v, i, j, k, v1, v2, v3, x:

		# No variables left? Then this point is
		# explicit.
		if nops(V) = 0 then
			# Is this point in the open interval?  Needs to be
			# decidedly IN or OUT, hence use EvalRel
			if EvalRel(p>a) and EvalRel(p<b) then
				return {p}:
			else
				return {}:
			end if:
		end if:

		# Clear the collection of solutions
		X := {}:

		# Look for a binary parameter
		i := 1:
		while i <= nops(V) and convert(V[i],string)[1..2] <> "_B" do
			i := i + 1:
		end do:

		# Did we find one?
		if i <= nops(V) then
			# Extract the parameter
			v := V[i]:
			V2 := subsop(i=NULL,copy(V)):

			# Perform the binary substitution
			X := X union GetParameterizedPointsInOpenInterval( subs(v=0,p), V2, a, b ):
			X := X union GetParameterizedPointsInOpenInterval( subs(v=1,p), V2, a, b ):
		else
			# We can only handle ONE non-binary parameter
			if nops(V) > 1 then
				error "too many parameters to handle: %1", p:
			end if:

			# Grab the parameter
			v := V[1]:

			# Acceptable parameter type?
			if convert(v,string)[1..2] <> "_Z" and
				convert(v,string)[1..3] <> "_NN" then
				error "unexpected parameter type: %1", v:
			end if:

			# TODO: This doesn't handle imaginary values!

			# Get bounds on this variable
			# in RealRange format
			v1 := {solve(p>a,v)}:
			v2 := {solve(p<b,v)}:

			# In case assumptions have been lost, build in
			# constraints based on the parameter type
			if convert(v,string)[1..2] = "_Z" then
				v3 := RealRange(-infinity,infinity):
			else
				v3 := RealRange(0,infinity):
			end if:

			# Run through all pairs of constraints and extract
			# possible solutions
			for i from 1 to nops(v1) do
				for j from 1 to nops(v2) do
					x := AndProp(v1[i],v2[j],v3):
					if not type(x,RealRange) then
						error "unable to solve for this parameter: %1", v:
					end if:

					# Get integer solutions in the current range
					x := GetIntegersInRealRange(x):

					# Substitute these into the original expression
					x := map(k->subs(v=k,p),x):

					# Merge with our current set of solutions
					X := X union x:
				end do:
			end do:
		end if:

		# Return the set of all solutions
		return X:
	end proc:


	# Given a univariate relation e in the variable x, finds
	# all solutions xi to e such that xi is in the open interval
	# (a,b)
	GetSolutionsInOpenInterval := proc( e, x, a, b )
		local X, XP, V, Vx, Vx1, Vx2, B, i, j, k, l, _EAS, _ETH, _EE:

		# Find possible branch points
		(_EAS,_EnvAllSolutions) := (_EnvAllSolutions,true):
		(_ETH,_EnvTryHard) := (_EnvTryHard,true):
		(_EE,_EnvExplicit) := (_EnvExplicit,true):
		X := {solve( e, x )}:
		if _SolutionsMayBeLost = true then
			error "solutions may be lost: %1", df=0:
		end if:
		_EnvAllSolutions := _EAS:
		_EnvTryHard := _ETH:
		_EnvExplicit := _EE:

		# Split into explicit and parameterized branch points
		(X,XP) := selectremove( i->evalb(nops(
			GetVars(i) minus GetVars(e))=0), X ):

		# Cull out-of-bounds and imaginary explicit branch points.
		# Points need to be conclusively IN or OUT, hence the use
		# of EvalRel
		X := select( i->EvalRel(i>a) and EvalRel(i<b) and Im(i) = 0, X ):

		# Do we have any parameterized branch points to deal with?
		if nops(XP) > 0 then
			for i from 1 to nops(XP) do
				V := GetVars(XP[i]) minus GetVars(e):
				X := X union GetParameterizedPointsInOpenInterval(XP[i],V,a,b):
			end do:
		end if:

		# Return the set of solutions
		return X:
	end proc:


	#################################################################
	# INVERSION CODE                                                #
	#################################################################


	# Returns ALL possible inverses g(y) to a continuous
	# function f(x).
	GetAllInverses := proc( f::EXPR, x::name, y::name )
		options remember:
		local _EAS, _ETH, _EE, G:

		# Ensure that the inverse variable does not
		# occur already in the function
		if has(f,y) then
			error "can't invert: %1 contains %2", f, y:
		end if:

		# Modify the solver behaviour
		(_EAS,_EnvAllSolutions) := (_EnvAllSolutions,true):
		(_ETH,_EnvTryHard) := (_EnvTryHard,true):
		(_EE,_EnvExplicit) := (_EnvExplicit,true):

		# Find all inverses
		G := {solve(f=y,x)}:

		# Revert back to original behaviour
		_EnvAllSolutions := _EAS:
		_EnvTryHard := _ETH:
		_EnvExplicit := _EE:

		# No inverses found?
		if nops(G) = 0 then
			error "unable to invert %1", f:
		end if:

		# HACK
		# Not in closed form?
		if has(G,RootOf) then
			error "unable to find closed form inverse to %1", f:
		end if:

		# Return the inverses
		return G:
	end proc:


	# Given a function f(x) defined on an open interval (a, b),
	# selects the appopriate inverse from a collection of (possibly
	# parameterized) candidates G.  Assumes f(x) is monotone on the
	# open interval (a,b)--may return incorrect results if this is not so.
	# NOTE: Can't use remember, because depends on assumptions in force!
	SelectInverse := proc( f, x, a, b, G, y )
		local i, j, n, X, fX, gX, VG, V, S:

		# Determine any parameters in each solution
		VG := convert(G,list):
		VG := map( i->GetVars(i) minus GetVars(f) minus {y}, VG ):

		# We need one sample per parameter in order
		# to determine the inverse
		n := max( op(map( i->nops(i), VG )), 1 ):

		# Overdetermine the system by 1
		n := n + 1:

		# Get n sample points in the open interval
		X := GetPointsInOpenInterval( a, b, n ):

		# Evaluate f at each point x in X
		try
			fX := simplify(eval(map(i->subs(x=i,f), X ))):
		catch :
			fX := simplify(eval(map(i->limit(f,x=i), X ))):
		end try:

		# Run through each possible inverse
		for i from 1 to nops(G) do
			# Evaluate the candidate inverse at each point f in fX
			gX := simplify(eval(map(j->subs(y=j,G[i]),fX))):

			# Is this a closed form inverse with no parameters?
			if VG[i] = {} then
				# Is this the inverse?
				gX := simplify(gX - X):
				if IsZeroVector(gX) then
					return G[i]:
				end if:
			# Otherwise, we have parameters to determine
			else
				# Create system of equations
				S := simplify(eval({seq(simplify(gX[j]=X[j]),j=1..nops(X))})):

				# Solve for free parameters
				# TODO: This is known not to work for things
				# like LambertW's, etc (ie: ln(x) + x^k), or
				# really complicated roots.
				V := solve(S,VG[i]):

				# HACK
				# Did we get closed form solutions for each parameter?
				if V <> NULL and nops(V) = nops(VG[i]) and not has(V,RootOf) then
					# Substitute the parameter values and simplify
					# using any assumptions.
					j := SimplifyExpr(subs(op(V),G[i])):

					# Return the calculated inverse
					return j:
				end if:

				# TODO: A numerical inverse finding technique.
				# Need to come up with a way to generate random values
				# for any parameters appearing in 'f', and then
				# generate a system of equations that only depend on
				# the parameters in the inverse (which are by necessity
				# integer valued)
				# Then need to try fsolve for those parameters, or,
				# failing that, try a finite number of random guesses.
			end if:
		end do:

		# If we get here, no appropriate inverse
		# could be found
		error "unable to choose inverse for %1 for %2 on (%3,%4)", f, x, a, b:
	end proc:


	# Given a continuous real-valued function f(x) defined on
	# the open-interval (a,b), and a candidate inverse g(y), tests
	# to see if the candidate is actually the inverse.
	TestInverse := proc( f, x, a, b, g, y )
		local n, X, fX, gfX, i:

		# Get some sample points
		n := 2:
		X := GetPointsInOpenInterval( a, b, n ):

		# Evaluate f at points x in X
		fX := simplify(eval(map(i->subs(x=i,f),X))):

		# Evaluate the candidate inverse at each point f in fX
		gfX := simplify(eval(map(i->subs(y=i,g),fX))):

		# Do they agree?
		gfX := simplify(gfX-X):
		return IsZeroVector(gfX):
	end proc:


	# Inverts a STRICTLY MONOTONE function f(x) defined on an open interval.
	# Returns the answer as the inverse.
	InvertMonotoneFunction := proc( f, x, a, b, y )
		options remember:
		local g, G:

		# Ensure that the inverse variable does not
		# occur already in the function
		if has(f,y) then
			error "can't invert: %1 contains %2", f, y:
		end if:
		
		# Calculate the simple inverse first
		G := {solve(f=y,x)}:

		# Did we get only one inverse?
		if nops(G) = 1 then
			G := G[1]:
			try
				# Test this inverse
				if TestInverse( f, x, a, b, G, y ) = true then
					# TODO: It is possible to get false positives
					# here: for example, the function we are inverting
					# has branch points on the interval, and the random
					# test points chosen by TestInverse happen to all
					# lie in one branch, which also happens to agree with
					# the canonical inverse calculated above.
					return G:
				end if:
			catch :
				# If testing fails, then gracefully continue
			end try:
		end if:

		# If we get here, the simple inverse wasn't good
		# enough, so let's get down to real work!

		# Get all possible inverses
		G := GetAllInverses( f, x, y ):

		# Choose the appropriate inverse
		g := SelectInverse( f, x, a, b, G, y ):

		# Return the applicable inverse
		return g:
	end proc:


	# Given a continuous non-constant function f(x) defined on
	# interval (a,b), calculates the general inverse wrt y.  Returns
	# the answer as a set of functions defined on the non-overlapping
	# open-intervals they apply to.
	# Can't use 'remember', because the answer may depend
	# on assumptions!
	GeneralInverseOnOpenInterval := proc( f, x, a, b, y )
		local G, g, i, j, ai, bi:

		# Ensure that the inverse variable does not
		# occur already in the function
		if has(f,y) then
			error "can't invert: %1 contains %2", f, y:
		end if:
		
		# Calculate the simple inverse first
		G := {solve(f=y,x)}:

		# Did we get only one inverse?
		if nops(G) = 1 then
			G := G[1]:
			try
				# Test this inverse
				if TestInverse( f, x, a, b, G, y ) = true then
					# TODO: It is possible to get false positives
					# here: for example, the function we are inverting
					# has branch points on the interval, and the random
					# test points chosen by TestInverse happen to all
					# lie in one branch, which also happens to agree with
					# the canonical inverse calculated above.
					return [[a,G,b]]:
				end if:
			catch :
				# If testing fails, then gracefully continue
			end try:
		end if:

		# If we get here, the simple inverse wasn't good
		# enough, so let's get down to real work!	

		# Determine any branch points
		g := GetSolutionsInOpenInterval( diff(f,x)=0, x, a, b ):

		# Sort these points: this MUST succeed otherwise
		# we are unable to determine the open intervals.
		g := convert( g, list ):
		g := sort( g, (i,j)->EvalRel(i<j) ):

		# Convert to non-overlapping open intervals
		g := [a,op(g),b]:
		for i from 1 to nops(g)-1 do
			g[i] := [g[i],g[i+1]]:
		end do:
		g := subsop(nops(g)=NULL,g):

		# Get all possible inverses
		G := GetAllInverses( f, x, y ):

		# Over each open interval, determine the appropriate inverse
		for i from 1 to nops(g) do
			(ai,bi) := op(g[i]):
			g[i] := [ai,SelectInverse(f,x,ai,bi,G,y),bi]:
		end do:

		# Return the general inverse
		return g:
	end proc:


	#################################################################
	# PWF OBJECT CODE                                               #
	#################################################################


	# Defines a PWF
	`_type/PWF` := proc( pwf, dim::nonnegint )
		if op(0,pwf) <> PWF then
			return false:
		end if:

		# Dimension provided?
		if nargs > 1 then
			if nops(op(2,pwf)) <> dim then
				return false:
			end if:
		end if:

		return true:
	end proc:


	# Pretty prints a PWF
	`_print/PWF` := proc( pl, vl, al, _pp::integer )
		local out, i, b, fb, x, x1, x2, vl2, inf, pp:

		# _pp passed?
		if nargs > 3 then
			pp := _pp:
		else
			# If one dimensional, print points
			#if nops(vl) = 1 then
			#	pp := 1:
			#else
			#	pp := 0:
			#end if:

			# FOR NOW, SHOW ALL POINTS UP TO 3 DIMENSIONS
			pp := 3:
		end if:

		# BASE CASE: Zero dimensional?
		if nops(vl) = 0 then
			return pl:
		end if:

		# Get main var and sub-varlist
		x := vl[1]:
		vl2 := subsop(1=NULL,vl):

		# Generate 'infinity' for our current dimension
		inf := infinity:
		for i from 1 to nops(vl2) do
			inf := [inf]:
		end do:

		# Special case: defined on whole real line?
		if nops(pl) = 1 then
			out := 'PIECEWISE'([`print/PWF`(pl[1],vl2,{},pp-1),'all'(x)]):
		# Special case: defined only at a point?
		elif nops(pl) = 4 and pl[1] = inf and pl[4] = inf then
			out := NULL:

			fb := `print/PWF`(pl[3],vl2,{},pp-1):
			out := out, [infinity,x<pl[2]]:
			out := out, [fb,x=pl[2]]:
			out := out, [infinity,pl[2]<x]:

			out := 'PIECEWISE'(out)
		# Defined on an interval
		else
			out := NULL:

			x2 := -infinity:
			for i from 1 to nops(pl) by 3 do
				# Get the bounding points
				if i < nops(pl) then
					(x1,x2) := (x2,pl[i+1]):
				else
					(x1,x2) := (x2,infinity):
				end if:

				fb := `print/PWF`(pl[i],vl2,{},pp-1):

				if x1 = -infinity then
					# (-inf,inf) : this can't happen, as
					# it's caught by the above special case
					if x2 = infinity then
						b := x:
					# (-inf,x2)
					else
						b := x<x2:
					end if:
				else
					# (x1,inf)
					if x2 = infinity then
						b := x1<x:
					# (x1,x2)
					else
						b := ``(x1<x) and ``(x<x2):
					end if:
				end if:

				out := out, [fb,b]:

				# Are we printing values at points?
				if pp > 0 and x2 <> infinity then
					fb := `print/PWF`(pl[i+2],vl2,{},pp-1):
					b := x = x2:
					out := out, [fb,b]:
				end if:
			end do:

			out := 'PIECEWISE'(out):
		end if:

		return out:
	end proc:


	# Simplifies a PWF given its components.
	`PWFPL/Simplify` := proc( _pl, vl )
		local pl, vl2, x, x1, x2, i, inf:

		# BASE CASE: Zero dimensional?
		if nops(vl) = 0 then
			return SimplifyExpr(_pl):
		end if:

		# Grab a copy of the piecelist
		pl := _pl:

		# Get the main var and sub-varlist
		x := op(1,vl):
		vl2 := subsop(1=NULL,vl):

		# Simplify the points (2 mod 3)
		for i from 2 to nops(pl) by 3 do
			pl[i] := SimplifyExpr(pl[i]):
		end do:

		# Merge repeated points
		i := 5:
		while i <= nops(pl) do
			# Successive points are the same?
			if pl[i-3] = pl[i] then
				# Remove the repeated point
				pl := subsop(i-3=NULL,i-2=NULL,i-1=NULL,pl):
			else
				# Advance to the next point
				i := i + 3:
			end if:
		end do:

		# Simplify each fi (1 mod 3)
		x2 := -infinity:
		for i from 1 to nops(pl) by 3 do
			x1 := x2:
			if i+1 < nops(pl) then
				x2 := pl[i+1]:
			else
				x2 := infinity:
			end if:
			AssumePush({x1<x,x<x2}):
			pl[i] := `PWFPL/Simplify`(pl[i],vl2):
			AssumePop():
		end do:

		# Generate infinity for this dimension
		inf := infinity:
		for i from 1 to nops(vl2) do
			inf := [inf]:
		end do:

		# TODO: This can't always be done; it's possible that
		# the function is defined as the same thing on either side
		# of a point, but defined as a limit in between.  We don't
		# want to merge in that case.
		# (fine and dandy in 1D, but messes things up in ND)
		# x1^2/x2 on x2>0 is an example

		# See if we can merge successive intervals
		#i := 4:
		#while i <= nops(pl) do
		#	# Successive intervals are the same?
		#	if (pl[i-3] = inf and pl[i-1] = inf and pl[i] = inf)
		#		or (pl[i-3] <> inf and pl[i-3] = pl[i]) then
		#		# Remove the point and right interval
		#		pl := subsop(i-2=NULL,i-1=NULL,i=NULL,pl):
		#	else
		#		# Advance to the next interval
		#		i := i + 3:
		#	end if:
		#end do:

		# Return the simplified piecelist
		return pl:
	end proc:


	# Attempts to simplify a PWF object as much
	# as possible (evaluates points, attempts to
	# merge intervals, etc).
	`PWF/Simplify` := proc( _pwf )
		local pl, vl, al:

		# Get the pwf bits
		(pl,vl,al) := op(_pwf):

		# Simplify the piecelist
		AssumePush(al):
		pl := `PWFPL/Simplify`(pl,vl):
		AssumePop():

		# Return the simplified PWF
		return PWF(pl,vl,al):
	end proc:


	# Converts a function (potentially a piecewise object) to
	# a piecelist.  Only supports one-dimension.  Does absolutely
	# no simplification.  Third (optional) argument specified
	# whether or not points should be filled in and continuity
	# tested.
	PiecewiseToPiecelist := proc ( _f, x, _b )
		local f, i, pl, x1, x2, f_i, felse, ptprev, t, b:

		# Get boolean
		if nargs > 2 then
			b := _b:
		else
			b := true:
		end if:

		# Grab a copy
		f := _f:

		# SPECIAL CASE: defined on whole real line
		if not (type(f,function) and op(0,f) = piecewise) then
			return [f]:
		end if:

		# Grab the piecewise operands
		f := [op(f)]:

		# Is there an explicit felse?
		i := nops(f):
		if i mod 2 = 1 then
			felse := f[i]:
			f := subsop(i=NULL,f):
		# No explicit felse
		else
			# Is there a 'something < x' at the end?
			# If so, there's an implied felse
			i := nops(f):
			if rhs(f[i-1]) = x then
				felse := f[i]:
				f := subsop(i-1=NULL,i=NULL,f):
			# Otherwise, there is no felse, so we use the default
			else
				felse := infinity:
			end if:
		end if:

		# Ensure valid conditions (x <, <=, = something)
		for i from 1 to nops(f)-1 by 2 do
			# Must be relations
			if not type(f[i],relation) then
				error "not a relation: %1", f[i]:
			end if:

			# Must have x as lhs
			if lhs(f[i]) <> x then
				error "invalid relation: %1", f[i]:
			end if:
		end do:

		# Convert to a piecelist
		ptprev := true:
		pl := NULL:
		f_i := felse:
		x2 := -infinity:
		for i from 1 to nops(f) by 2 do
			# Get point
			(x1,x2) := (x2,rhs(f[i])):

			# Point?
			if type(f[i],`=`) then
				# Point previous?
				if ptprev then
					pl := pl, f_i:
				end if:

				pl := pl, x2, f[i+1]:

				ptprev := true:
			# Interval and point?
			elif type(f[i],`<=`) then
				f_i := f[i+1]:

				# Interval previous?
				if not ptprev then
					pl := pl, x1, f_i:
				end if:

				pl := pl, f_i:
				pl := pl, x2, f_i:

				ptprev := true:
			# Interval only?
			elif type(f[i],`<`) then
				f_i := f[i+1]:

				# Interval previous?
				if not ptprev then
					pl := pl, x1, f_i:
				end if:

				pl := pl, f_i:

				ptprev := false:
			end if:
		end do:

		# Ended on an interval?
		if not ptprev then
			pl := pl, x2, felse:
		end if:
		pl := pl, felse:

		# Complete the PWFPL
		pl := [pl]:

		# Should we bother with the points?
		if b then
			# Fill in points to ensure lsc
			for i from 2 to nops(pl) by 3 do
				# Undefined left interval?
				if pl[i-1] = infinity then
					# Undefined right interval?
					if pl[i+2] = infinity then
						# Ensure no var appears
						if has(pl[i+1],x) then
							pl[i+1] := subs(x=pl[i],pl[i+1]):
						end if:
					# Defined right interval?
					else
						# Take a limit from the right interval
						pl[i+1] := limit(pl[i+2],x=pl[i],right) assuming op(AssumeCur):
					end if:
				# Defined left interval?
				else
					# Take a limit from the left interval
					pl[i+1] := limit(pl[i-1],x=pl[i],left) assuming op(AssumeCur):

					# Defined right interval?
					if pl[i+2] <> infinity then
						# Take a limit from the right interval
						t := limit(pl[i+2],x=pl[i],right) assuming op(AssumeCur):

						# Do the left and right limits mismatch?
						if not AreEqual(t,pl[i+1]) then
							error "discontinuous at %1", pl[i]:
						end if:
					end if:
				end if:
			end do:
		end if:

		return pl:
	end proc:


	# Recursively converts a piecewise function
	# into a piecelist
	PiecewiseToPwfPl := proc(
		f,
		vl,
		rl
		)
		local i, x, a, b, vl2, rl2, pl, pw:

		# BASE CASE
		if nops(vl) = 0 then
			return f:
		end if:

		# Get the current variable
		x := vl[1]:

		# Get the bounds on the current variable
		a := simplify(rl[1][1]) assuming op(AssumeCur):
		b := simplify(rl[1][2]) assuming op(AssumeCur):

		# TODO: Might want to make sure these are in order!

		# Get the sublists
		vl2 := subsop(1=NULL,vl):
		rl2 := subsop(1=NULL,rl):

		# Convert to piecewise wrt current variable
		if a <> -infinity then
			if b <> infinity then
				pw := piecewise( a<=x and x<=b, f, infinity ):
			else
				pw := piecewise( a<=x, f, infinity ):
			end if:
		else
			if b <> infinity then
				pw := piecewise( x<=b, f, infinity ):
			else
				pw := f:
			end if:
		end if:

		pw := convert(pw,piecewise,x) assuming op(AssumeCur):

		# Convert to a piecelist
		pl := PiecewiseToPiecelist( pw, x ):

		# Recursively process the sub pls
		b := -infinity:
		for i from 1 to nops(pl) do
			# Process sub-intervals
			if i mod 3 = 1 then
				# Get interval end-points
				a := b:
				if i+1 <= nops(pl) then
					b := pl[i+1]:
				else
					b := infinity:
				end if:

				AssumePush({a<=x,x<=b}):
				pl[i] := PiecewiseToPwfPl(pl[i],vl2,rl2):
				AssumePop():
			# Process sub points
			elif i mod 3 = 0 then
				AssumePush({x=b}):
				pl[i] := subs(x=b,pl[i]):
				pl[i] := PiecewiseToPwfPl(pl[i],vl2,rl2):
				AssumePop():
			end if:
		end do:

		return pl:
	end proc:


	# Converts a function to a PWF object.
	`_convert/PWF` := proc(
		_f::EXPR, _x, _al
		#_x::Or(name,`=`(name,`..`),list(Or(name,`=`(name,`..`))))
		#_al::set(ASSUMPTION)
		)
		local i, pl, vl, rl, al, pwf, f, R, S:

		# Grab a copy of the function
		f := _f:

		# Main variable provided?
		if nargs > 1 then
			if type(_x,list) then
				vl := _x:
			else
				vl := [_x]:
			end if:
		# If not, try to guess it!
		else
			# Get the variables that appear in the function
			vl := GetVars(f):

			# If none do, make one up!
			if vl = {} then
				vl := [x]:
			elif nops(vl) = 1 then
				vl := [vl[1]]:
			# If there's more than one, bail out!
			else
				error "unable to determine variable name from %1", x:
			end if:
		end if:

		# Extract variables and ranges that they apply on
		rl := vl:
		for i from 1 to nops(vl) do
			if type(vl[i],`=`) then
				rl[i] := [op(rhs(vl[i]))]:
				vl[i] := lhs(vl[i]):
			else
				rl[i] := [-infinity,infinity]:
			end if:
		end do:

		# Could be that we have expressions such
		# as abs(x[i]) in the expressions for x[j], j > i.
		# The following is a hack to fix that
		S := 0:
		for i from 1 to nops(rl) do
			if not has(rl[i][1],infinity) then
				S := S + (_C || i || 1) * rl[i][1]:
			end if:
			if not has(rl[i][2],infinity) then
				S := S + (_C || i || 2) * rl[i][2]:
			end if:
		end do:
		S := _PwfConst * S:

		# Reset the assumption stack
		AssumeResetStack():

		# Assumptions provided?
		if nargs > 2 then
			al := _al:
			
			# Push the provided assumptions
			AssumePush(al):
		else
			al := {}:
		end if:

		# Convert to a piecelist
		pl := PiecewiseToPwfPl( S + f, vl, rl ):

		# Remove the _PwfConst terms used in the
		# transformation
		pl := subs(_PwfConst = 0, pl):

		# Reset the assumption stack again
		AssumeResetStack():

		# Make a new pwf
		pwf := Assume(PWF(pl,vl,{}),op(al)):

		# Simplify the newly minted PWF
		pwf := `PWF/Simplify`(pwf):

		return pwf:
	end proc:


	# Adds two elements together, taking care of
	# obvious simplifications involving infinities.
	PwfBopAdd := proc( e1, e2 )
		if e1 = infinity or e2 = infinity then
			return infinity:
		end if:
		if e1 = -infinity or e2 = -infinity then
			error "-infinity not allowed":
		end if:
		return e1 + e2:
	end proc:


	# Multiplies two elements together, taking care of
	# obvious simplifications involving infinities.
	PwfBopMult := proc( e1, e2 )
		if e1 = infinity or e2 = infinity then
			return infinity:
		end if:
		if e1 = -infinity or e2 = -infinity then
			error "-infinity not allowed":
		end if:
		return e1*e2:
	end proc:


	# Powers to elements together, taking care of
	# obvious simplifications.
	PwfBopPow := proc( e1, e2 )
		if e1 = infinity or e2 = infinity then
			return infinity:
		end if:
		if e1 = -infinity or e2 = -infinity then
			error "-infinity not allowed":
		end if:
		return e1^e2:
	end proc:


	# Applies a binary operand to two piecelists.  They must
	# be in the same variable order!
	PwfApplyBopPlPl := proc( pl1, pl2, vl, opname )
		local x, vl2, pl, x1, x2, i1, i2, xl, xr:

		# BASE CASE: Zero dim?
		if nops(vl) = 0 then
			return opname(pl1,pl2):
		end if:

		# Grab the main variable and sub-varlist
		x := op(1,vl):
		vl2 := subsop(1=NULL,vl):

		(i1,i2) := (1,1):
		if i1 < nops(pl1) then
			x1 := pl1[i1+1]:
		else
			x1 := infinity:
		end if:
		if i2 < nops(pl2) then
			x2 := pl2[i2+1]:
		else
			x2 := infinity:
		end if:

		# Step through the intervals
		pl := NULL:
		xr := -infinity:
		while xr <> infinity do
			# Output the current interval
			pl := pl, PwfApplyBopPlPl(pl1[i1],pl2[i2],vl2,opname):

			# Get the limiting point, and advance
			# the appropriate index/indices
			if AreEqual(x1,x2) then
				(xl,xr) := (xr,x1):
				if xr <> infinity then
					pl := pl, xr, PwfApplyBopPlPl(pl1[i1+2],pl2[i2+2],vl2,opname):
				end if:
				i1 := i1 + 3:
				i2 := i2 + 3:
				if i1 < nops(pl1) then
					x1 := pl1[i1+1]:
				else
					x1 := infinity:
				end if:
				if i2 < nops(pl2) then
					x2 := pl1[i2+1]:
				else
					x2 := infinity:
				end if:
			elif EvalRel(x1<x2) then
				(xl,xr) := (xr,x1):
				if xr <> infinity then
					# TODO: Is the subs ok here?
					pl := pl, xr, PwfApplyBopPlPl(pl1[i1+2],
						subs(x=xr,pl2[i2]),vl2,opname):
				end if:
				i1 := i1 + 3:
				if i1 < nops(pl1) then
					x1 := pl1[i1+1]:
				else
					x1 := infinity:
				end if:
			else
				(xl,xr) := (xr,x2):
				if x2 <> infinity then
					# TODO: Is the subs ok here?
					pl := pl, xr, PwfApplyBopPlPl(
						subs(x=xr,pl1[i1]),pl2[i2+2],vl2,opname):
				end if:
				i2 := i2 + 3:
				if i2 < nops(pl2) then
					x2 := pl2[i2+1]:
				else
					x2 := infinity:
				end if:
			end if:
		end do:
		
		# Return the result
		pl := [pl]:
		return pl:
	end proc:


	# Applies a binary operand to a PWF and an expression
	PwfApplyBopPlExpr := proc( _pl, vl, e, opname )
		local pl, vl2, x, i:

		# BASE CASE: Zero dim?
		if nops(vl) = 0 then
			return opname(_pl,e):
		end if:

		# Make a copy of the pl
		pl := _pl:

		# Get the main var and sub-varlist
		x := op(1,vl):
		vl2 := subsop(1=NULL,vl):

		# Walk through the piecelist (skipping points)
		for i from 1 to nops(pl) do
			# Point?
			if i mod 3 = 2 then
				# Skip it!
				next:
			end if:

			# Apply the operator to this piece
			pl[i] := PwfApplyBopPlExpr(pl[i],vl2,e,opname):
		end do:

		# Return the result
		return pl:
	end proc:


	# Applies a binary operand to a PWF and an expression
	PwfApplyBopExprPl := proc( e, _pl, vl, opname )
		local pl, vl2, x, i:

		# BASE CASE: Zero dim?
		if nops(vl) = 0 then
			return opname(e,_pl):
		end if:

		# Make a copy of the pl
		pl := _pl:

		# Get the main var and sub-varlist
		x := op(1,vl):
		vl2 := subsop(1=NULL,vl):

		# Walk through the piecelist (skipping points)
		for i from 1 to nops(pl) do
			# Point?
			if i mod 3 = 2 then
				# Skip it!
				next:
			end if:

			# Apply the operator to this piece
			pl[i] := PwfApplyBopExprPl(e,pl[i],vl2,opname):
		end do:

		# Return the result
		return pl:
	end proc:


	# Applies a binary operator to two PWFs or a PWF and
	# an expression.
	PwfApplyBop := proc( _e1, _e2, opname )
		local pl1, pl2, vl, al, e1, e2:

		# Grab copies
		(e1,e2) := (_e1,_e2):

		# e1 eligible for conversion?
		if not type(e1,PWF) then
			e1 := convert(e1,piecewise):
			if op(0,e1) = piecewise then
				e1 := convert(e1,PWF):
			end if:
		end if:

		# e2 eligible for conversion?
		if not type(e2,PWF) then
			e2 := convert(e2,piecewise):
			if op(0,e2) = piecewise then
				e2 := convert(e2,PWF):
			end if:
		end if:

		if type(e1,PWF) then
			# pwf,pwf
			if type(e2,PWF) then
				# Ensure the PWFs are sync'd
				# Could also upsample lower dimensional
				# PWFs.
				# TODO

				# Extract components (merge assumptions
				# from both PWFs)
				(pl1,vl,al) := op(e1):
				pl2 := op(1,e2):
				al := al union op(3,e2):

				# Apply the operator
				AssumePush(al):
				pl1 := PwfApplyBopPlPl(pl1,pl2,vl,opname):
				AssumePop():

				# Return the result
				return PWF(pl1,vl,al):
			# pwf,expr
			elif type(e2,EXPR) then
				# Extract components
				(pl1,vl,al) := op(e1):

				# Apply the operator
				AssumePush(al):
				pl1 := PwfApplyBopPlExpr(pl1,vl,e2,opname):
				AssumePop():

				# Return the result
				return PWF(pl1,vl,al):
			end if:
		elif type(e2,EXPR) then
			# expr,pwf
			if type(e2,PWF) then
				# Extract components
				(pl2,vl,al) := op(e2):

				# Apply the operator
				AssumePush(al):
				pl2 := PwfApplyBopExprPl(e1,pl2,vl,opname):
				AssumePop():

				# Return the result
				return PWF(pl2,vl,al):
			end if:
		end if:

		# If we get here, one or both of the
		# operands were invalid types!
		error "expect at least one PWF and at most one EXPR":
	end proc:


	# Applies a unary operand to a PL
	PwfApplyUopPl := proc( _pl, vl, opname )
		local x, pl, vl2, i:

		# BASE CASE: Zero dim?
		if nops(vl) = 0 then
			return opname(_pl):
		end if:

		# Grab a copy of the pl
		pl := _pl:

		# Get main var and sub-varlist
		x := op(1,vl):
		vl2 := subsop(1=NULL,vl):

		# Walk through the piecelist
		for i from 1 to nops(pl) do
			# Skip points
			if i mod 3 = 2 then
				next:
			end if:

			# Recursively process sub-piecelists
			pl[i] := PwfApplyUopPl(pl[i],vl2,opname):
		end do:

		# Return the piecelist
		return pl:
	end proc:


	# Applies a binary operand to a PWF
	PwfApplyUop := proc( pwf, opname )
		local pl, vl, al:

		# Extract the components
		(pl,vl,al) := op(pwf):

		# Apply the unary op to the piecelist
		AssumePush(al):
		pl := PwfApplyUopPl( pl, vl, al, opname ):
		AssumePop():

		# Return the PWF
		return PWF(pl,vl,al):
	end proc:


	# Attempts to simplify an expression containing
	# a piecewisefunc
	`_simplify/PWF` := proc( e )
		local res, i, uop, bop:

		bop := NULL:
		uop := NULL:

		# Reset the assumption stack
		AssumeResetStack():

		# Determine the base operator type
		if type(e,`+`) then
			bop := PwfBopAdd:
		elif type(e,`*`) then
			bop := PwfBopMult:
		elif type(e,`^`) then
			bop := PwfBopPow:
		elif type(e,function) and nops(e) = 1 then
			uop := op(0,e):
		end if:

		# Apply any binary operators
		if bop <> NULL then
			# TODO: Check variable order and dimension.
			# Try to make consistent.  If that fails, return
			# unevaluated.
			res := simplify(op(1,e)):
			for i from 2 to nops(e) do
				res := PwfApplyBop(
					res,
					simplify(op(i,e)),
					bop ):
			end do:
		# Apply any unary operators
		elif uop <> NULL then
			res := PwfApplyUop( simplify(op(1,e)), uop ):
		# We don't know what to do!
		else
			res := e:
		end if:

		# If the result is a PWF
		if type(res,PWF) then
			# Simplify the result
			res := `PWF/Simplify`(res):
		end if:

		return res:
	end proc:


	#################################################################
	# SD OBJECT CODE                                                #
	#################################################################


	# Defines an SD object.
	`_type/SD` := proc( sd )
		return evalb(op(0,sd)=SD):
	end proc:


	# Pretty prints a PWF object
	`_print/SD` := proc( pl, vl, al )
		local out, i, b, fb, x, x1, x2, vl2:

                # BASE CASE: Zero dimensional?
                if nops(vl) = 0 then
                        return pl:
                end if:

		# Get main var and sub-varlist
		x := vl[1]:
                vl2 := subsop(1=NULL,vl):

		# Special case: defined on whole real line?
		if nops(pl) = 1 then
			out := pl[1]:
                # Special case?

		# Defined on multiple intervals
		else
			out := NULL:

			x2 := -infinity:
			for i from 1 to nops(pl) by 3 do
				# Get the bounding points
				if i < nops(pl) then
					(x1,x2) := (x2,pl[i+1]):
				else
					(x1,x2) := (x2,infinity):
				end if:

				fb := pl[i]:

				if x1 = -infinity then
					# (-inf,inf) : this can't happen, as
					# it's caught by the above special case
					if x2 = infinity then
						b := x:
					# (-inf,x2)
					else
						b := x<x2:
					end if:
				else
					# (x1,inf)
					if x2 = infinity then
						b := x1<x:
					# (x1,x2)
					else
						b := ``(x1<x) and ``(x<x2):
					end if:
				end if:

				# Output the interval
				out := out, [`print/SD`(fb,vl2,al),b]:

				# Have a point to output?
				if x2 <> infinity then
					fb := pl[i+2]:
					b := x=x2:
					out := out, [`print/SD`(fb,vl2,al),b]:
				end if:
			end do:

			out := 'PIECEWISE'(out):
		end if:

		return out:
	end proc:


	# Simplifies an SD object
	`SD/Simplify` := proc( sd )
		local pl, vl, al, i, x1, x2:

		# Get the constituents
		(pl,vl,al) := op(sd):

		# Merge identical successive points
		i := 5:
		while i < nops(pl) do
			if pl[i] = pl[i-3] then
				pl := subsop(i-3=NULL,i-2=NULL,i-1=NULL,pl):
			else
				i := i + 3:
			end if:
		end do:

		# Simplify each fi (1 mod 3)
		x2 := -infinity:
		for i from 1 to nops(pl) by 3 do
			x1 := x2:
			if i+1 < nops(pl) then
				x2 := pl[i+1]:
			else
				x2 := infinity:
			end if:

			# A closed interval?  If the endpoints
			# are the same, collapse them.
			if nops(pl[i]) = 2 then
				if pl[i][1] = pl[i][2] then
					pl[i] := {pl[i][1]}:
				end if:
			end if:

			# Singleton set?
			if nops(pl[i]) = 1 then
				AssumePush({x1<x,x<x2}):
				pl[i] := SimplifyExpr(pl[i]):
				AssumePop():
			end if:
		end do:

		# See if we can merge successive intervals
		i := 4:
		while i <= nops(pl) do
			# Successive intervals are the same?
			if type(pl[i-3],set) and type(pl[i-1],set) and type(pl[i],set) and
				pl[i-3] = pl[i] then
				# Remove the point and right interval
				pl := subsop(i-2=NULL,i-1=NULL,i=NULL,pl):
			else
				# Advance to the next interval
				i := i + 3:
			end if:
		end do:

		# Reconstruct the simplified SD
		return SD(pl,vl,al):
	end proc:


	# Converts functions/piecewise objects to an SD object
	`_convert/SD` := proc( _f::EXPR, _x::name, _al::set(ASSUMPTION) )
		local sd, x, pl, vl, al, f, i, df1, df2:

		# Grab a copy of the function
		f := _f:

		# Main variable provided?
		if nargs > 1 then
			x := _x:
		# If not, try to guess it!
		else
			# Get the variables that appear in the function
			x := GetVars(f) minus {undefined}:

			# If none do, make one up!
			if x = {} then
				x := 'x':
			# If there's only one, use it!
			elif nops(x) = 1 then
				x := op(x):
			# If there's more than one, bail out!
			else
				error "unable to determine variable name from %1", x:
			end if:
		end if:

		# Reset the assumption stack
		AssumeResetStack():

		# Convert the provided function to piecewise
		f := convert(f,piecewise,x) assuming op(AssumeCur):

		# Convert to a piecelist (don't do lsc/continuity checking)
		pl := PiecewiseToPiecelist( f, x, false ):
		vl := [x]:

		# Assumptions provided?
		if nargs > 2 then
			al := _al:
		else
			al := {}:
		end if:

		# Convert intervals to SD format
		for i from 1 to nops(pl) by 3 do
			# Undefined?
			if pl[i] = infinity or pl[i] = undefined then
				pl[i] := {}:
			# Defined?
			else
				pl[i] := {pl[i]}:
			end if:
		end do:

		# Fill in the points
		pl := SdFillPoints(pl,vl):

		# Make a new SD
		sd := Assume(SD(pl,vl,{}),op(al)):

		# Simplify the newly minted SD
		sd := `SD/Simplify`(sd):

		return sd:
	end proc:


	# Adds two elements, applying SD relevant simplifications
	SdBopAdd := proc( e1, e2 )
		if e1 = undefined or e2 = undefined then
			return undefined:
		elif e1 = infinity or e2 = infinity then
			return infinity:
		else
			return e1 + e2:
		end if:
	end proc:

	
	# Multiplies two elements, applying SD relevant simplifications
	SdBopMult := proc( e1, e2 )
		if e1 = undefined or e2 = undefined then
			return undefined:
		elif e1 = infinity or e2 = infinity then
			return infinity:
		else
			return e1 * e2:
		end if:
	end proc:

	
	# Powers two elements, applying SD relevant simplifications
	SdBopPow := proc( e1, e2 )
		if e1 = undefined or e2 = undefined then
			return undefined:
		else
			return e1 ^ e2:
		end if:
	end proc:


	# Fills in the points on an SD PL
	SdFillPoints := proc( _pl, vl )
		local pl, x, i, df1, df2:

		pl := _pl:
		x := vl[1]:

		# Fill in points
		for i from 2 to nops(pl) by 3 do
			# Left undefined?
			if pl[i-1] = {} then
				# Right undefined?
				if pl[i+2] = {} then
					if pl[i+1] = {} or pl[i+1] = infinity or pl[i+1] = undefined then
						pl[i+1] := {}
					else
						pl[i+1] := [-infinity,infinity]:
					end if:
				# Right defined?
				else
					df2 := SimplifyExpr(limit(pl[i+2][1],x=pl[i],right)):
					if df2 = -infinity then
						pl[i+1] := {}:
					else
						pl[i+1] := [-infinity,df2]:
					end if:
				end if:
			# Left defined?
			else
				# Right undefined?
				if pl[i+2] = {} then
					df1 := SimplifyExpr(limit(pl[i-1][1],x=pl[i],left)):
					if df1 = infinity then
						pl[i+1] := {}:
					else
						pl[i+1] := [df1,infinity]:
					end if:
				# Right defined?
				else
					df1 := SimplifyExpr(limit(pl[i-1][1],x=pl[i],left)):
					df2 := SimplifyExpr(limit(pl[i+2][1],x=pl[i],right)):
					if AreEqual(df1,df2) then
						pl[i+1] := {df1}:
					else
						pl[i+1] := [df1,df2]:
					end if:
				end if:
			end if:
		end do:

		return pl:
	end proc:


	# Applies a binary operand to two SDs, or an SD and something else.	
	SdApplyBop := proc( e1, e2, opname )
		local sd, pl1, pl2, vl1, vl2, al1, al2, i, SdCleanPl:

		# Parse the operands

		if not type(e1,SD) and not type(e2,SD) then
			error "expect at least one SD":
		end if:

		if type(e1,SD) then
			(pl1,vl1,al1) := op(e1):
		else
			pl1 := NULL:
		end if:

		if type(e2,SD) then
			(pl2,vl2,al2) := op(e2):
		else
			pl2 := NULL:
		end if:

		if pl1 = NULL then
			pl1 := `_convert/SD`(e1,vl2[1]):
			(pl1,vl1,al1) := op(pl1):
		end if:

		if pl2 = NULL then
			pl2 := `_convert/SD`(e2,vl1[1]):
			(pl2,vl2,al2) := op(pl2):
		end if:

		# Ensure variables match!
		if vl1 <> vl2 then
			error "variable mismatch":
		end if:

		SdCleanPl := proc( _pl )
			local pl, i:
			pl := _pl:
			for i from 1 to nops(pl) do
				# Point?
				if i mod 3 = 2 then
					next:
				# Interval?
				elif i mod 3 = 1 then
					if pl[i] = {} then
						pl[i] := undefined:
					else
						pl[i] := op(pl[i]):
					end if:
				# Point value?
				else
					if pl[i] = {} then
						pl[i] := undefined:
					else
						pl[i] := 0:
					end if:
				end if:
			end do:
			return pl:
		end proc:

		# Clean the PLs (so they're PWF compatible)
		pl1 := SdCleanPl(pl1):
		pl2 := SdCleanPl(pl2):

		# Merge the assumptions
		al1 := al1 union al2:

		# Apply them
		AssumePush(al1):

		# Synchronize the piecelists
		pl1 := PwfApplyBopPlPl(pl1,pl2,vl1,`SCAT/BOPMERGE`):

		# Apply the operator
		for i from 1 to nops(pl1) do
			# Skip points
			if i mod 3 = 2 then
				next:
			end if:
			pl1[i] := {opname(op(pl1[i]))}:
		end do:

		# Fill in the points
		pl1 := SdFillPoints(pl1,vl1):

		# Pop the assumptions
		AssumePop():

		# Make the SD
		sd := Assume(SD(pl1,vl1,{}),op(al1)):

		return sd:
	end proc:

	
	# Applies a unary operand to an SD
	SdApplyUop := proc( _sd, opname )
		local sd, pl, vl, al, i:

		(pl,vl,al) := op(_sd):

		# Push the assumptions
		AssumePush(al):

		# Apply the operand to the intervals
		for i from 1 to nops(pl) by 3 do
			if nops(pl[i]) = 1 then
				pl[i] := {opname(pl[i][1])}:
			end if:
		end do:

		# Fill in the points
		pl := SdFillPoints(pl,vl):

		# Pop the assumptions
		AssumePop():

		# Make the SD
		sd := Assume(SD(pl,vl,{}),op(al)):

		return sd:
	end proc:

	
	# Simplifies expression containing SDs.
	`_simplify/SD`  := proc( e )
		local res, i, uop, bop:

		bop := NULL:
		uop := NULL:

		# Determine the base operator type
		if type(e,`+`) then
			bop := SdBopAdd:
		elif type(e,`*`) then
			bop := SdBopMult:
		elif type(e,`^`) then
			bop := SdBopPow:
		elif type(e,function) and nops(e) = 1 then
			uop := op(0,e):
		end if:

		# Apply any binary operators
		if bop <> NULL then
			res := simplify(op(1,e)):
			for i from 2 to nops(e) do
				res := SdApplyBop(
					res,
					simplify(op(i,e)),
					bop ):
			end do:
		# Apply any unary operators
		elif uop <> NULL then
			res := SdApplyUop( simplify(op(1,e)), uop ):
		# We don't know what to do!
		else
			res := e:
		end if:

		# If the result is a SD
		if type(res,SD) then
			# Simplify the result
			res := `SD/Simplify`(res):
		end if:

		return res:
	end proc:


	#################################################################
	# EVALUATION CODE                                               #
	#################################################################


	# Evaluates a piecelist at the given set of points
	`PWFPL/Eval` := proc( pl, xl )
		local pl2, xl2, i, x, xv:

		# Grab the first element, and the rest of the list
		x := op(1,xl[1]):
		xl2 := subsop(1=NULL,xl):

		# Is a variable specified?
		if type(xl[1],`=`) then
			# Find which interval this value falls within
			i := 2:
			xv := op(2,xl[1]):
			while i < nops(pl) and AlwaysTrue(pl[i] < xv) do
				i := i + 3:
			end do:

			# Equality with point?
			if i < nops(pl) and AreEqual(pl[i],xv) then
				pl2 := pl[i+1]:
			# In open interval?
			else
				pl2 := simplify(subs(x=xv,pl[i-1])):
			end if:

			# More dimensions to go?
			if nops(xl2) > 0 then
				pl2 := `PWFPL/Eval`(pl2,xl2):
			end if:
		# Otherwise, this is a 'free' variable
		else
			# Process each piece along this variable
			pl2 := pl:
			if nops(xl2) > 0 then
				for i from 1 to nops(pl2) do
					# Skip variables
					if i mod 3 = 2 then
						next:
					end if:

					# Recursively process this sub-piecelist
					pl2[i] := `PWFPL/Eval`(pl2[i],xl2):
				end do:
			end if:
		end if:

		return pl2:
	end proc:


	# Evalutes a PWF object at the given points
	# This is simply a wrapper to PWFPL/Eval.
	`PWF/Eval` := proc( pwf, xl )
		local pl, vl, al:

		# Get constituents
		(pl,vl,al) := op(pwf):

		# Push the assumptions
		AssumePush(al):

		# Do the evaluation
		pl := `PWFPL/Eval`(pl,xl):

		# Pop the assumptions
		AssumePop():

		# Result is a PWFPL?
		if type(pl,list) then
			vl := remove(type,xl,`=`):
			pl := PWF(pl,vl,{}):
			pl := Assume(pl,op(al)):
			pl := `PWF/Simplify`(pl):
		end if:

		return pl:
	end proc:


	# Evaluates a sub-differential
	`SD/Eval` := proc( sd, xl )
		local e, pl, vl, al, x, xv, i:

		# Is the variable specified?
		if type(xl[1],`=`) then
			# Get the constituents
			(pl,vl,al) := op(sd):

			# Push the assumptions
			AssumePush(al):

			# Get the variable and value
			(x,xv) := op(xl[1]):

			# Find which interval this value falls within
			i := 2:
			while i < nops(pl) and AlwaysTrue(pl[i] < xv) do
				i := i + 3:
			end do:

			# Equality with point?
			if i < nops(pl) and AreEqual(pl[i],xv) then
				e := pl[i+1]:
			# In open interval?
			else
				e := simplify(subs(x=xv,pl[i-1])):
			end if:

			# Pop the assumptions
			AssumePop():

			return e:
		else
			# Do nothing!
			return:
		end if:
	end proc:


	# Evaluates PWF or SD objects.
	Eval := proc( f )
		local pl, vl, al, xl, t, i, j:

		# Ensure sufficient arguments
		if nargs = 1 then
			error "insufficient arguments":
		end if:

		# Is f a PWF?
		if type(f,PWF) then
			(pl,vl,al) := op(f):
		# SD object?
		elif type(f,SD) then
			(pl,vl,al) := op(f):
		# Anything else?
		else
			error "expect SD or PWF, received: %1", f:
		end if:

		# Extract any var=val type arguments
		xl := vl:
		t := subsop(1=NULL,[args]):
		i := 1:
		while i <= nops(t) do
			if type(t[i],`=`) then
				if member(lhs(t[i]),vl,'j') then
					if xl[j] = lhs(t[i]) then
						xl[j] := t[i]:
						t := subsop(i=NULL,t):
					else
						error "duplicate variable: %1", lhs(t[i]):
					end if:
				else
					error "unknown variable: %1", lhs(t[i]):
				end if:
			else
				i := i + 1:
			end if:
		end do:

		# Fill in any other variables
		i := 1:
		j := 1:
		while i <= nops(xl) and j <= nops(t) do
			if not type(xl[i],`=`) then
				if not type(t[j],EXPR) then
					error "invalid argument: %1", t[j]:
				end if:
				xl[i] := xl[i]=t[j]:
				j := j + 1:
			end if:
			i := i + 1:
		end do:

		# Propagate any backwards references
		for i from 2 to nops(xl) do
			if type(xl[i],`=`) then
				t := select(type,[op(1..i-1,xl)],`=`):
				xl[i] := simplify(subs(op(t),xl[i])):
			end if:
		end do:

		# Ensure no forward references
		for i from 1 to nops(xl) do
			if type(xl[i],`=`) and has(rhs(xl[i]),vl) then
				error "invalid argument: %1", xl[i]:
			end if:
		end do:

		# Reset the assumption stack
		AssumeResetStack():

		# Is f a PWF?
		if type(f,PWF) then
			return `PWF/Eval`(f,xl):
		# SD object?
		elif type(f,SD) then
			return `SD/Eval`(f,xl):
		end if:
	end proc:
	protect(Eval):


	#################################################################
	# PLOTTING CODE                                                 #
	#################################################################


	# Validates a range, and return the end-points
	# as floats.
	EvaluateRange := proc( r )
		local x1, x2:

		# Get the endpoints
		(x1,x2) := op(r):

		# Empty range?
		if x1 = x2 then
			error "empty range: %1", r:
		end if:

		# Evaluate them
		(x1,x2) := (Re(evalf(x1)),Re(evalf(x2))):

		# Don't allow Float(infinity)'s
		(x1,x2) := op(subs(Float(infinity)=infinity,Float(-infinity)=-infinity,[x1,x2])):

		# Floats?
		if not type(x1,numeric) or not type(x2,numeric) then
			error "invalid range: %1", r:
		end if:

		# Equal?
		if x1 = x2 then
			error "empty range: %1", r:
		# Misordered?
		elif x1 > x2 then
			error "out of order range: %1", r:
		end if:

		# Return the evaluated range
		return x1,x2:
	end proc:


	# Plots a one-dimensional PWF
	`PWF/Plot1D` := proc( pwf, r )
		local pl, vl, al, i, j, P, x, x1, x2, xl, xr, xa, xb, opt:

		# Get the range
		(x1,x2) := EvaluateRange(r):

		# Get the constituents
		(pl,vl,al) := op(pwf):

		# Get the variable
		x := op(vl):

		# Ensure we have no unbound variables
		P := GetVars(pl) minus {x}:
		if nops(P) <> 0 then
			error "unbound variables: %1", P:
		end if:

		# Get any additional options
		if nargs > 2 then
			opt := [args[3..nargs]]:
		else
			opt := []:
		end if:

		# Walk through the intervals
		P := NULL:
		xr := -infinity:
		for i from 1 to nops(pl) by 3 do
			# Get the end-points of the current interval
			xl := xr:
			if i < nops(pl) then
				xr := evalf(pl[i+1]):
			else
				xr := infinity:
			end if:

			# Is this interval undefined?
			if pl[i] = infinity then
				next:
			end if:

			# Is this interval left of the plotting interval?
			if xr <= x1 then
				next:
			end if:

			# Is this interval right of the plotting interval?
			if xl >= x2 then
				next:
			end if:

			# Get the overlap
			xa := max(x1,xl):
			xb := min(x2,xr):

			# Plot this interval over this range
			P := P, plot(pl[i],x=xa..xb,op(opt)):
		end do:

		# Walk through the points
		for i from 2 to nops(pl) by 3 do
			# Is this point in the interval?
			if evalf(pl[i]) >= x1 and evalf(pl[i]) <= x2 then
				try
					pl[i+1] := evalf(subs(x=pl[i],pl[i+1])):
					plots[pointplot]([[pl[i],pl[i+1]]],
						symbol=circle,color=black,op(opt)):
					P := P, %:
				catch:
					# If the evaluation of the point fails,
					# this is no reason for the rest of the
					# plot to fail, so absorb this error.
				end try:
			end if:
		end do:

		# Look for the view option
		for i from 1 to nops(opt) do
			if type(opt[i],`=`) and op(1,opt[i]) = view then
				break:
			end if:
		end do:

		# No view option?
		if i > nops(opt) then
			# Make a default one
			opt := [op(opt),view=[DEFAULT,DEFAULT]]:
		# Had view option?
		else
			# If not list type, then make into list type
			# view=...?
			if not type(op(2,opt[i]),list) then
				opt[i] := view=[DEFAULT,op(2,opt[i])]:
			end if:
		end if:

		# Get the view option list
		j := op(2,opt[i]):

		# Default x view?
		if j[1] = DEFAULT then
			j[1] := x1..x2:
		end if:

		# Store the modified view option
		opt[i] := view=j:

		# Produce the compound plot
		P := plots[display](P,op(opt)):

		# Return it
		return P:
	end proc:


	# Plots a two-dimensional PWF
	`PWF/Plot2D` := proc( pwf, rx, ry )
		local pl, vl, al, i, j, P,
			x, x1, x2, xl, xr, xa, xb,
			y, y1, y2, yl, yr, ya, yb,
			Y1, Y2, opt, f, Rp, Rx, Ry, Rs:

		# Get the ranges (x range MUST be numeric, y
		# range may be function of x)
		(x1,x2) := EvaluateRange(rx):
		(y1,y2) := op(ry):

		# Get the constituents
		(pl,vl,al) := op(pwf):

		# Get the variables
		(x,y) := op(vl):

		# y1 parameterized?
		if not type(y1,numeric) then
			# Valid parameterization?
			P := GetVars(y1) minus {x}:
			if nops(P) > 0 then
				error "unbound variables: %1", P:
			end if:
		end if:

		# y2 parameterized?
		if not type(y2,numeric) then
			# Valid parameterization?
			P := GetVars(y2) minus {x}:
			if nops(P) > 0 then
				error "unbound variables: %1", P:
			end if:
		end if:

		# Ensure we have no unbound variables
		P := GetVars(pl) minus {x,y}:
		if nops(P) > 0 then
			error "unbound variables: %1", P:
		end if:

		# Get any additional options
		if nargs > 3 then
			opt := [args[4..nargs]]:
		else
			opt := []:
		end if:

		# PLOTTING STUFF

		# Convert to regions
		Rs := PwfPlToRegions(pl):

		# Split into points, x lines, y lines and surfaces
		Rx,Rs := selectremove(i->i[1]=i[2],Rs):
		Rp,Rx := selectremove(i->i[3]=i[4],Rx):
		Ry,Rs := selectremove(i->i[3]=i[4],Rs):

		# Start with blank plot
		P := NULL:

		# Have surfaces?
		if nops(Rs) > 0 then
			for i from 1 to nops(Rs) do
				xa,xb,ya,yb,f := op(Rs[i]):
				
				# Have x overlap?
				if xa < x2 and xb > x1 then
					# Get the overlap
					(xl,xr) := max(xa,x1),min(xb,x2):
				end if:

				# Simple y vars?
				if not has(ya,x) and not has(yb,x) then
					# Have y overlap?
					if ya < y2 and yb > y1 then
						# Get the overlap
						(yl,yr) := max(ya,y1),min(yb,y2):

						P := P, plot3d( f, x=xl..xr, y=yl..yr, op(opt) ):
					end if:
				else
					if has(ya,x) then
						yl := ya:
					else
						yl := minimize(yb,x=xl..xr):
					end if:
					if has(yb,x) then
						yr := yb:
					else
						yr := maximize(ya,x=xl..xr):
					end if:

					P := P, plot3d( f, x=xl..xr, y=yl..yr,
						view=[x1..x2,y1..y2,DEFAULT], op(opt) ):
				end if:
			end do:			
		# Have lines?
		elif nops(Rx) + nops(Ry) > 0 then
			# Have x lines?
			if nops(Rx) > 0 then
				# TODO
			end if:

			# Have y lines?
			if nops(Ry) > 0 then
				# TODO
			end if:
		# Have points?
		elif nops(Rp) > 0 then
			# TODO
		# Have nothing?!
		else
			# This should never happen!
			error "empty plot":
		end if:

		if false then

		# Below is view adjustment stuff, necessary to make the
		# plot3d view command work with display.

		# Get the view option index
		for i from 1 to nops(opt) do
			if type(opt[i],`=`) and op(1,opt[i]) = view then
				break:
			end if:
		end do:

		# No view option?
		if i > nops(opt) then
			# Make a default one
			opt := [op(opt),view=[DEFAULT,DEFAULT,DEFAULT]]:
		# Had view option?
		else
			# If not list type, then make into list type
			# view=...?
			if not type(op(2,opt[i]),list) then
				opt[i] := view=[DEFAULT,DEFAULT,op(2,opt[i])]:
			end if:
		end if:

		# Get the view option list
		j := op(2,opt[i]):

		# Is the x view left to default?
		if j[1] = DEFAULT then
			j[1] := x1..x2:
		end if:

		# Is the y view left to default?
		if j[2] = DEFAULT then
			# Get minimum y value
			Y1 := min(minimize(y1,x=x1..x2),
					minimize(y2,x=x1..x2)):

			# Get maximum y value
			Y2 := max(maximize(y1,x=x1..x2),
					maximize(y2,x=x1..x2)):

			j[2] := Y1..Y2:
		end if:

		# Plug the modified view back into the options
		opt[i] := view=j:

		end if:

		# Produce the compound plot
		P := plots[display](P,op(opt)):

		# Return it
		return P:
	end proc:


	# Plots an SD object
	`SD/Plot` := proc( sd, r )
		local pl, vl, al, x1, x2, xl, xr, xa, xb, opt, x, P, i, j,
			y1, y2, yr, ymin, ymax, vi:

		# Get the range
		(x1,x2) := EvaluateRange(r):

		# Get the constituents
		(pl,vl,al) := op(sd):
		pl := evalf(pl):
		pl := subs(Float(infinity)=infinity,Float(-infinity)=-infinity,pl):

		# Get the variable
		x := op(vl):

		# Ensure we have no unbound variables
		P := GetVars(pl) minus {x}:
		if nops(P) <> 0 then
			error "unbound variables: %1", P:
		end if:

		# Get any additional options
		if nargs > 2 then
			opt := [args[3..nargs]]:
		else
			opt := []:
		end if:

		# Find the view option, if any
		for vi from 1 to nops(opt) do
			# Found view=... option?
			if type(opt[vi],`=`) and op(1,opt[vi])=view then
				# Is view=...?  Turn into view=[...,...]
				if not type(op(2,opt[vi]),list) then
					opt[vi] := view=[DEFAULT,op(2,opt[vi])]:
				end if:
				break:
			end if:
		end do:

		# None found?  Make one!
		if vi > nops(opt) then
			opt := [op(opt),view=[DEFAULT,DEFAULT]]:
		end if:

		# Run through the intervals
		P := NULL:
		xr := -infinity:
		(y1,y2) := (0,0):
		for i from 1 to nops(pl) by 3 do
			# Get the endpoints
			xl := xr:
			if i < nops(pl) then
				xr := pl[i+1]:
			else
				xr := infinity:
			end if:

			# Defined?
			if pl[i] <> {} then
				# Left of the interval?
				if xr < x1 then
					next:
				end if:

				# Right of the interval?
				if x2 < xl then
					next:
				end if:

				# Get the overlap
				xa := max(x1,xl):
				xb := min(x2,xr):

				# First segment?
				if P = NULL then
					# Remember the left end-point
					y1 := limit(pl[i][1],x=xa,right):
				end if:

				# Remember the right end-point
				y2 := limit(pl[i][1],x=xb,left):

				# Plot this segment
				P := P, plot(pl[i][1],x=xa..xb,op(opt)):

				# Is the right end-point in the plotting range?
				if xb < x2 then
					# Is the right end-point an interval?
					if type(pl[i+2],list) then
						# Is it finite?
						if pl[i+2][2] <> infinity then
							# Plot it!						
							P := P, plot([[xb,pl[i+2][1]],
								[xb,pl[i+2][2]]],op(opt)):
						end if:
					end if:
				end if:
			end if:
		end do:

		# Calculate the y range (this will be
		# the default length of the +inf and -inf bars)
		yr := abs(y2-y1):
		if yr = 0 then
			yr := 1:
		end if:

		# Start with end-points as y bounds
		(ymin,ymax) := (y1,y2):

		# Are there bounds on the y view?
		j := op(2,opt[vi])[2]:

		# Look for a [-infinity,...] point
		for i from 3 to nops(pl) by 3 do
			if type(pl[i],list) and pl[i][1] = -infinity then
				# Fall within viewable range?
				if pl[i-1] >= x1 and pl[i-1] <= x2 then
					# Bound on y?
					if type(j,`..`) then
						if op(1,j) < y1 then
							ymin := op(1,j):
						end if:
					else
						ymin := y1-yr:
					end if:

					# Plot it!
					P := P, plot([[pl[i-1],ymin],[pl[i-1],y1]],op(opt)):
				end if:
			end if:
		end do:

		# Look for a [...,infinity] point
		for i from 3 to nops(pl) by 3 do
			if type(pl[i],list) and pl[i][2] = infinity then
				# Fall within viewable range?
				if pl[i-1] >= x1 and pl[i-1] <= x2 then
					# Bound on y?
					if type(j,`..`) then
						if op(2,j) > y2 then
							ymax := op(2,j):
						end if:
					else
						ymax := y2+yr:
					end if:

					# Plot it!
					P := P, plot([[pl[i-1],y2],[pl[i-1],ymax]],op(opt)):
				end if:
			end if:
		end do:

		# Retrieve view command
		j := op(2,opt[vi]):

		# Default x range?
		if j[1] = DEFAULT then
			j[1] := x1..x2:
		end if:

		# Default y range?
		if j[2] = DEFAULT then
			#j[2] := ymin..ymax:
		end if:

		# Store view command
		opt[vi] := view=j:

		# Make the compound plot
		P := plots[display](P,op(opt)):

		# Return the plot!
		return P:
	end proc:


	# Plots a 1D PWF, 2D PWF, or 1D SubDiff, choosing the appropriate
	# plotting engine.  Valid formats:
	# Plot( PWF(1), x=a..b, options... )
	# Plot( PWF(1), a..b, options... )
	# Plot( SD, x=a..b, options... )
	# Plot( SD, a..b, options... )
	# Plot( PWF(2), x=a..b, y=c..d, options... )
	# Plot( PWF(2), x=a..b, c..d, options... )
	# Plot( PWF(2), a..b, y=c..d, options... )
	# Plot( PWF(2), a..b, c..d, options... )
	Plot := proc( f )
		local opt, rx, ry, x, y:

		# Reset the assumption stack
		AssumeResetStack():

		# Set or list?
		if type(f,set) or type(f,list) then
			# Get the options
			if nargs > 1 then
				opt := args[2..nargs]:
			else
				opt := NULL:
			end if:

			# Plot each one individually
			ry := NULL:
			for rx from 1 to nops(f) do
				ry := ry, Plot(f[rx],opt):
			end do:

			# Return a compount plot
			return plots[display]([ry]):
		end if:

		# 1D PWF?
		if type(f,PWF(1)) then
			# Sufficient arguments?
			if nargs < 2 then
				error "insufficient arguments":
			end if:

			# Get the range
			rx := args[2]:

			# Get any options
			if nargs > 2 then
				opt := args[3..nargs]:
			else
				opt := NULL:
			end if:

			# x=a..b?
			if type(rx,`=`) then
				# Matching variable?
				if op(1,rx) <> op([2,1],f) then
					error "unknown variable: %1", op(1,rx):
				end if:
				rx := op(2,rx):
			end if:

			# Do the plot!
			return `PWF/Plot1D`(f,rx,opt):
		# 2D PWF?
		elif type(f,PWF(2)) then
			# Sufficient arguments?
			if nargs < 3 then
				error "insufficient arguments":
			end if:

			# Get the ranges
			rx := args[2]:
			ry := args[3]:

			# Get any options
			if nargs > 3 then
				opt := args[4..nargs]:
			else
				opt := NULL:
			end if:

			# Get the variables
			(x,y) := op(op(2,f)):

			# Both `=`?
			if type(rx,`=`) and type(ry,`=`) then
				if op(1,rx) = x and op(1,ry) = y then
					(rx,ry) := (op(2,rx),op(2,ry)):
				elif op(1,rx) = y and op(1,ry) = x then
					(rx,ry) := (op(2,ry),op(2,rx)):
				else
					error "unknown variable":
				end if:
			# rx `=?
			elif type(rx,`=`) then
				if op(1,rx) = x then
					(rx,ry) := (op(2,rx),ry):
				elif op(1,rx) = y then
					(rx,ry) := (ry,op(2,rx)):
				else
					error "unknown variable: %1", op(1,rx):
				end if:
			# ry `=`?
			elif type(ry,`=`) then
				if op(1,ry) = x then
					(rx,ry) := (op(2,ry),rx):
				elif op(1,ry) = y then
					(rx,ry) := (rx,op(2,ry)):
				else
					error "unknown variable: %1", op(1,ry):
				end if:
			end if:

			# Do the plot!
			return `PWF/Plot2D`(f,rx,ry,opt):
		# Subdiff?
		elif type(f,SD) then
			# Sufficient arguments?
			if nargs < 2 then
				error "insufficient arguments":
			end if:

			# Get the range
			rx := args[2]:

			# Get any options
			if nargs > 2 then
				opt := args[3..nargs]:
			else
				opt := NULL:
			end if:

			# x=a..b?
			if type(rx,`=`) then
				# Matching variable?
				if op(1,rx) <> op([2,1],f) then
					error "unknown variable: %1", op(1,rx):
				end if:
				rx := op(2,rx):
			end if:

			# Do the plot!
			return `SD/Plot`(f,rx,opt):
		end if:
		
		# If we get here, things are bad!
		error "expect PWF(1), PWF(2) or SD: %1", f:
	end proc:
	protect(Plot):


	#################################################################
	# EQUALITY TESTING CODE                                         #
	#################################################################


	# Compares two PWFPLs.  Returns true, false or FAIL.
	`PWFPL/Equal` := proc( _pl1, _pl2, vl )
		local pl, pl1, pl2, ret, i, j, x, vl2, X, Xr, x1, x2, t, pl1t, pl2t:

		# BASE CASE: Zero dim?  Then we're comparing
		# two expressions for equality.
		if nops(vl) = 0 then
			try
				ret := AreEqual(_pl1,_pl2):
			catch :
				ret := FAIL:
			end try:
			return ret:
		end if:

		# Synchronize point locations in piecelists
		pl := PwfApplyBopPlPl(_pl1,_pl2,vl,`SCAT/BOPMERGE`):

		# Get the synchronized pl1
		i := (x,y) -> x:
		pl1 := eval(subs(`SCAT/BOPMERGE`=i,pl)):

		# Get the synchronized pl2
		i := (x,y) -> y:
		pl2 := eval(subs(`SCAT/BOPMERGE`=i,pl)):

		# Get main var, and sub-varlist
		x := vl[1]:
		vl2 := subsop(1=NULL,vl):

		# Walk through the intervals
		x2 := -infinity:
		for i from 1 to nops(pl) by 3 do
			# Get variable boundaries
			x1 := x2:
			if i < nops(pl) then
				x2 := pl[i+1]:
			else
				x2 := infinity:
			end if:

			# Do a simple compare first!
			if pl1[i] = pl2[i] then
				next:
			end if:

			# Grab some points in the open interval
			t := GetPointsInOpenInterval(x1,x2,3):
			
			# Test along each of these points
			for j from 1 to nops(t) do
				pl1t := subs(x=t[j],pl1[i]):
				pl2t := subs(x=t[j],pl2[i]):
				
				# Compare these intervals
				ret := `PWFPL/Equal`(pl1t,pl2t,vl2):
				if ret <> true then
					return ret:
				end if:
			end do:
		end do:

		# If we get here, they compared equal!
		return true:
	end proc:


	# Compares two PWFs.  Returns true, false, or FAIL.
	`PWF/Equal` := proc( pwf1, pwf2 )
		local pl1, pl2, vl1, vl2, al1, al2, b:

		# Get constituents
		(pl1,vl1,al1) := op(pwf1):
		(pl2,vl2,al2) := op(pwf2):

		# Different variables?
		if convert(vl1,set) <> convert(vl2,set) then
			return false:
		end if:

		# TODO: Could try to put variables in same order
		if vl1 <> vl2 then
			return FAIL:
		end if:

		# Compare the piecelists
		AssumePush(al1 union al2):
		b := `PWFPL/Equal`(pl1,pl2,vl1):
		AssumePop():

		return b:
	end proc:


	# Compare two SDs.  Returns true, false or FAIL.
	`SD/Equal` := proc( sd1, sd2 )
		local pl1, pl2, vl1, vl2, SDPL2PWFPL:

		# Get constituents
		(pl1,vl1) := op(1..2,sd1):
		(pl2,vl2) := op(1..2,sd2):

		# Different main variables?
		if vl1 <> vl2 then
			return false:
		end if:

		# Converts an SDPL to a pseudo-PWFPL
		SDPL2PWFPL := proc( _pl )
			local i, pl:
			pl := _pl:
			for i from 1 to nops(pl) do
				# Interval?
				if i mod 3 = 1 then
					if pl[i] = {} then
						pl[i] := infinity:
					else
						pl[i] := pl[i][1]:
					end if:
				# Point?
				elif i mod 3 = 0 then
					pl[i] := infinity:
				end if:
			end do:
		end proc:

		# Convert to PWFPLs for comparing
		pl1 := SDPL2PWFPL(pl1):
		pl2 := SDPL2PWFPL(pl2):

		# Defer the real work to PWFPL/Equal
		return `PWFPL/Equal`(pl1,pl2,vl1):
	end proc:


	# Compare two PWFs or SDs.  Returns true, false or FAIL.
	Equal := proc( f1, f2 )
		# Reset the assumption stack
		AssumeResetStack():

		# PWFs?
		if type(f1,PWF) and type(f2,PWF) then
			return `PWF/Equal`(f1,f2):
		# SDs?
		elif type(f1,SD) and type(f2,SD) then
			return `SD/Equal`(f1,f2):
		# Invalid input?
		else
			error "expect PWFs or SDs, received: %1, %2", f1, f2:
		end if:
	end proc:
	protect(Equal):


	#################################################################
	# LIMIT TAKING CODE                                             #
	#################################################################

	
	# Takes the limit of a PWFPL.
	`PWFPL/Limit` := proc( _pl, vl, xe, dir )
		local pl, i, x, x1, x2, vl2:

		# BASE CASE: Zero dim?
		if nops(vl) = 0 then
			return limit(_pl,xe,dir) assuming op(AssumeCur):
		end if:

		# Grab copy of PWFPL
		pl := _pl:

		# Get main var and subvarlist
		x := vl[1]:
		vl2 := subsop(1=NULL,vl):

		# Walk through the points
		for i from 2 to nops(pl) by 3 do
			pl[i] := limit(pl[i],xe,dir) assuming op(AssumeCur):
			pl[i+1] := `PWFPL/Limit`(pl[i+1],vl2,xe,dir) assuming op(AssumeCur):
		end do:

		# Collapse any points that have become equal
		i := 5:
		while i < nops(pl) do
			if EvalRel(pl[i-3] = pl[i]) then
				pl := subsop(i-3=NULL,i-2=NULL,i-1=NULL,pl):
			else
				i := i + 3:
			end if:
		end do:

		# Walk through the sub PWFPLs
		x2 := -infinity:
		for i from 1 to nops(pl) by 3 do
			# Get the constraints
			x1 := x2:
			if i < nops(pl) then
				x2 := pl[i+1]:
			else
				x2 := infinity:
			end if:

			# Push the constraints on this variable
			AssumePush({x1<x,x<x2}):

			# Perform the recursive simplification
			pl[i] := `PWFPL/Limit`(pl[i],vl2,xe,dir) assuming op(AssumeCur):

			# Pop the constraints
			AssumePop():
		end do:

		return pl:
	end proc:


	# Takes the limit of a PWF
	`PWF/Limit` := proc( _pwf, xe, dir )
		local pwf, pl, vl, al:

		# Get the constituents
		(pl,vl,al) := op(_pwf):

		# Push the assumptions
		AssumePush(al):

		# Take the limit
		pl := `PWFPL/Limit`(pl,vl,xe,dir):

		# Pop the assumptions
		AssumePop():

		# Reconstruct the PWF
		pwf := PWF(pl,vl,al):

		# Simplify the PWF
		pwf := `PWF/Simplify`(pwf):

		return pwf:
	end proc:


	# Takes the limit of an SD object
	`SD/Limit` := proc( _sd, xe, dir )
		local sd, pl, vl, al, i:

		# Get the constituents
		(pl,vl,al) := op(_sd):

		# Push the assumptions
		AssumePush(al):

		# Walk through the SDPL
		for i from 1 to nops(pl) do
			 # Interval?
			 if i mod 3 = 1 then
				# Singleton set?
				if nops(pl[i]) = 1 then
					pl[i] := {limit(pl[i][1],xe,dir)}:

					# Collapsed to infinity?
					if has(pl[i],infinity) then
						# Is actually the empty set here!
						pl[i] := {}:
					end if:
				end if:
			 # Point?
			 elif i mod 3 = 2 then
				pl[i] := limit(pl[i],xe,dir):
			 # Value at point?
			 else
				# Singleton set?
				if nops(pl[i]) = 1 then
					pl[i] := {limit(pl[i][1],xe,dir)}:

					# Collapsed to infinity?
					if has(pl[i],infinity) then
						# Is actually the empty set here!
						pl[i] := {}:
					end if:
				# Closed range?
				elif nops(pl[i]) = 2 then
					pl[i][1] := limit(pl[i][1],xe,dir):
					pl[i][2] := limit(pl[i][2],xe,dir):

					# Collapsed to singleton?
					if pl[i][1] = pl[i][2] then
						pl[i] := {pl[i][1]}:
					end if:
				end if:
			 end if:
		end do:

		# Pop the assumptions
		AssumePop():

		# Glue together the SD
		sd := SD(pl,vl,al):

		# Simplify the SD
		sd := `SD/Simplify`(sd):

		return sd:
	end proc:


	# Takes the limit of a PWF or SD
	Limit := proc( f, xe, _dir )
		local dir, l, vl, al:

		# Validate input type
		if not type(f,PWF) and not type(f,SD) then
			error "expect PWF or SD, received %1", f:
		end if:

		# Validate relation
		if not type(xe,`=`) or not type(op(1,xe),name) then
			error "expect name=expr, received %1", xe:
		end if:

		# Get direction
		if nargs > 2 then
			dir := _dir:
		else
			dir := real:
		end if:

		# Validate direction
		if not has({left,right,real},dir) then
			error "expect left, right or real, received %1", dir:
		end if:

		# Get the variables and assumptions
		vl := op(2,f):
		al := op(3,f):

		# Ensure no main variable appears in lhs
		if has(vl,op(1,xe)) then
			error "can't take limit of main variable, use Eval instead":
		end if:

		# Reset the assumption stack
		AssumeResetStack():

		# PWF?
		if type(f,PWF) then
			# Take the limit
			l := `PWF/Limit`(f,xe,dir):
		# SD?
		else
			# Take the limit
			l := `SD/Limit`(f,xe,dir):
		end if:

		# Trim the assumptions, as some may no longer be
		# valid
		l := AssumeTrim(l):

		# Return the evaluated object
		return l:
	end proc:
	protect(Limit):


	#################################################################
	# SUBSTITUTION CODE                                             #
	#################################################################


	# Performs substitutions (in a safe manner, followed by
	# simplification) into a PWF or SD object.
	# Accepts the same syntax as subs(...).
	# Will allow generic expr1=expr2 substitutions, as long as
	# expr1 does NOT contain a main variable.  Will allow variable
	# change substitutions of the form name1=name2, where name1 is
	# a main variable.
	Subs := proc()
		local f, vl, S, i:

		# Nothing to do?
		if nargs <= 1 then
			return args:
		end if:

		# Get the last argument
		f := args[nargs]:

		# Ensure is PWF or SD
		if not type(f,Or(PWF,SD)) then
			error "expect PWF or SD, received %1", f:
		end if:

		# Get the varlist
		vl := op(2,f):

		# Get the substitutions, and check them
		S := NULL:
		for i from 1 to nargs-1 do
			# Is `=`?
			if not type(args[i],`=`) then
				error "expect `=`, received %1", args[i]:
			end if:

			# lhs has var from vl?
			if has(args[i],vl) then
				# Ensure is name=name
				if type(op(1,args[i]),name) and
					type(op(2,args[i]),name) then
					# Valid substitution
					S := S, args[i]:
				else
					error "invalid substitution: %1", args[i]:
				end if:
			# lhs has no var from vl?
			else
				# Accept this substitution
				S := S, args[i]:
			end if:
		end do:

		# Perform the substitution
		f := subs(S,f):

		# Reset (and trim) the assumptions list, because some
		# of them may have been invalidated by the substitution.
		f := Assume(f,op(op(3,f))):

		# Reset the assumption stack
		AssumeResetStack():

		# Do the appropriate simplification
		if type(f,PWF) then
			f := `PWF/Simplify`(f):
		else
			f := `SD/Simplify`(f):
		end if:

		return f:
	end proc:
	protect(Subs):


	#################################################################
	# SUBDIFFERENTIATION CODE                                       #
	#################################################################


	# Calculates the subdifferential of a 1D PWFPL, returning
	# an SDPL.
	`PWFPL/SubDiff` := proc( pl, vl )
		options remember:
		local sdpl, i, dl, dr, x:

		# Get the variable
		x := vl[1]:

		# Grab a copy
		sdpl := pl:

		# Compute the derivatives along the intervals
		for i from 1 to nops(sdpl) by 3 do
			if sdpl[i] = infinity then
				sdpl[i] := {}:
			else
				sdpl[i] := {SimplifyExpr(diff(sdpl[i],x))}:
			end if:
		end do:

		# Fill in the points
		sdpl := SdFillPoints(sdpl,vl):

		return sdpl:
	end proc:


	# Calculates a 1D subdiff.  Is simply a wrapper to PWFPL/SubDiff
	# that also resets the assumption stack.
	SubDiff := proc( pwf::PWF(1) )
		options remember:
		local pl, vl, al, sdpl:

		# Reset the assumption stack
		AssumeResetStack():

		# Get the components
		(pl,vl,al) := op(pwf):

		# Push the assumptions
		AssumePush(al):

		# Calculate the subdiff
		sdpl := `PWFPL/SubDiff`(pl,vl):

		# Pop the assumptions
		AssumePop(al):

		# Return the conjugate
		return Assume(SD(sdpl,vl,{}),op(al)):
	end proc:
	protect(SubDiff):


	#################################################################
	# INTEGRATION CODE                                              #
	#################################################################


	# Integrates an SD object, returning a PWF.
	`SD/Integrate` := proc( sd )
		local pwf, pl, vl, al, x, x1, x2, i, t:

		# Get the constituents
		(pl,vl,al) := op(sd):
		x := vl[1]:

		# Push the assumptions
		AssumePush(al):

		# Walk through the intervals
		x2 := -infinity:
		for i from 1 to nops(pl) by 3 do
			x1 := x2:
			if i < nops(pl) then
				x2 := pl[i+1]:
			else
				x2 := infinity:
			end if:

			# Singleton set?
			if nops(pl[i]) = 1 then
				AssumePush({x1<x,x<x2}):
				pl[i] := int(pl[i][1],x):
				AssumePop():

				# Success?
				if type(pl[i],function) and op(0,pl[i]) = int then
					error "unable to compute %1", pl[i]:
				end if:
			# Undefined?
			else
				pl[i] := infinity:
			end if:
		end do:

		# Fill in points, and align neighboring intervals
		for i from 2 to nops(pl) by 3 do
			# Undefined left interval?
			if pl[i-1] = infinity then
				# Undefined right interval?
				if pl[i+2] = infinity then
					# Subdiff undefined?
					if pl[i+1] = {} then
						pl[i+1] := infinity:
					else
						pl[i+1] := 0:
					end if:
				# Defined right interval?
				else
					# Take a limit from the right interval
					pl[i+1] := limit(pl[i+2],x=pl[i],right):
				end if:
			# Defined left interval?
			else
				# Take a limit from the left interval
				pl[i+1] := limit(pl[i-1],x=pl[i],left):

				# Defined right interval?
				if pl[i+2] <> infinity then
					# Take a limit from the right interval
					t := limit(pl[i+2],x=pl[i],right):

					# Do the left and right limits mismatch?
					if t <> pl[i+1] then
						# Align the right interval to the left one
						pl[i+2] := pl[i+2] - t + pl[i+1]:
					end if:
				end if:
			end if:
		end do:

		# Pop the assumptions
		AssumePop():

		# Make a pwf
		pwf := Assume(PWF(pl,vl,{}),op(al)):

		# Simplify the resulting PWF
		pwf := `PWF/Simplify`(pwf):
		
		return pwf:	
	end proc:


	# Integrates an SD, returning a PWF
	Integ := proc( sd::SD )
		# Reset the assumption stack
		AssumeResetStack():

		# Calculate the integral
		return `SD/Integrate`(sd):
	end proc:
	protect(Integ):


	#################################################################
	# 1D CONJUGATION CODE                                           #
	#################################################################


	# Calculates the fenchel conjugate of an (assumed to be
	# convex and continuous) one-dimensional PWFPL.
	`PWFPL/Conj1D` := proc( _pl, vl, vl2 )
		local pl, dp, i, j, x, y, G, g, t, u, v, x1, x2, n:

		# Get the variables
		pl := _pl:
		x := vl[1]:
		y := vl2[1]:

		# Calculate the sub-differential, and grab the
		# piecelist
		dp := `PWFPL/SubDiff`(pl,vl):

		# NOTE: When adding to G, always end with a point,
		# even if that point is infinity!

		# Walk through the intervals
		G := NULL:
		x2 := -infinity:
		for i from 1 to nops(pl) by 3 do
			# Get the interval bounds
			x1 := x2:
			if i < nops(pl) then
				x2 := pl[i+1]:
			else
				x2 := infinity:
			end if:

			# Is the subdiff defined?
			if dp[i] <> {} then
				# Is the interval subdiff a function of x?
				if has(dp[i],x) then		
					# Get the general inverse
					t := GeneralInverseOnOpenInterval(
						dp[i][1], x, x1, x2, y ):
					
					# Calculate the conjugate over each chunk of the
					# inverse
					for j from 1 to nops(t) do
						t[j][2] := SimplifyExpr(subs(x=t[j][2],x*y-pl[i])):
					end do:

					# Get the end-points for each piece
					u := SimplifyExpr(limit(dp[i][1],x=t[1][1],right)):
					t[1][1] := u:
					for j from 1 to nops(t) do
						v := SimplifyExpr(limit(dp[i][1],x=t[j][3],left)):
						t[j][3] := v:
						if j > 1 then
							t[j][1] := t[j-1][3]:
						end if:
					end do:

					# Simplify each piece based on variable constraints
					for j from 1 to nops(t) do
						AssumePush({t[j][1]<y,y<t[j][3]}):
						t[j][2] := SimplifyExpr(t[j][2]):
						AssumePop():
					end do:

					# List is empty?
					if G = NULL then
						u := t[1][1]:
						# First point doesn't occur at -infinity?
						if u <> -infinity then
							# Output the first point
							v := SimplifyExpr(limit(t[1][2],y=u,right)):
							G := infinity,u,v:
						end if:
					end if:

					# Walk through the intervals
					for j from 1 to nops(t) do
						# Output interval and right end-point
						v := SimplifyExpr(limit(t[j][2],y=t[j][3],left)):
						G := G, t[j][2], t[j][3], v:
					end do:
				# The interval subdiff is constant
				else
					# Is the list empty?
					if G = NULL then
						# Get the point at which the conjugate is defined
						u := dp[i][1]:

						# This point is finite?
						if u <> infinity then
							t := SimplifyExpr(subs(x=0,-pl[i])):
							G := infinity, u, t:
						end if:
					end if:
				end if:
			end if: # Subdiff is defined

			# Does a point exist, and if so, is the subdiff
			# defined at that point?
			if i < nops(pl) and dp[i+2] <> {} then
				# Is the point an interval?
				if type(dp[i+2],list) then
					# HACK: This avoids problems with input PWFs
					# that aren't actually LSC, by taking the appropriate
					# limit to determine the function value at x=pl[i+1].
					if pl[i+2] = infinity then
						if pl[i] = infinity then
							j := limit(pl[i+3],x=x2,right):
						else
							j := limit(pl[i],x=x2,left):
						end if:
					else
						j := pl[i+2]:
					end if:

					# Grab the interval endpoints
					(u,v) := op(dp[i+2]):

					# Find the conjugate
					AssumePush({u<y,y<v}):
					t := SimplifyExpr(subs(x=x2,x*y-j)):
					AssumePop():

					# List is empty?
					if G = NULL then
						# First point not at -infinity?
						if u <> -infinity then
							# Output point
							j := SimplifyExpr(limit(t,y=u,right)):
							G := infinity, u, j:
						end if:
					end if:

					# Output this interval and the right end-point
					j := SimplifyExpr(limit(t,y=v,left)):
					G := G, t, v, j:
				# This point is a singleton
				else
					# Is the list empty?
					if G = NULL then
						# Get the point at which the conjugate is defined
						u := dp[i][1]:

						# This point is finite?
						if u <> infinity then
							t := SimplifyExpr(u*x2-pl[i+2]):
							G := infinity, u, t:
						end if:
					end if:
				end if:
			end if: # Subdiff is defined
		end do:

		# Convert sequence to list
		G := [G]:

		# SPECIAL CASE: Empty list?
		n := nops(G):
		if n = 0 then
			G := [infinity]:
		# Non-empty (contains at least "point,interval,point")
		else
			# Ends with point at infinity?
			if G[n-1] = infinity then
				# Remove that point!
				G := subsop(n-1=NULL,n=NULL,G):
			# No point at infinity?
			else
				# Fill in the last interval
				G := [op(G),infinity]:
			end if:
		end if:

		# Return the conjugated PWFPL
		return G:
	end proc:


	#################################################################
	# PIVOTING CODE                                                 #
	#################################################################


	# Converts a PWFPL to a list of REGIONS.
	# REGIONS are odd lengthed lists of the form
	# [x1_1,x1_2,x2_1,x2_2,....,xn_1,x_n2,f(x_1,x_2,...,x_n)]
	PwfPlToRegions := proc( pl )
		local R, R2, i, j, x1, x2:

		# BASE CASE: Zero dimensional?
		if not type(pl,list) then
			if pl = infinity then
				return []:
			else
				return [[pl]]:
			end if:
		end if:

		# Clear the region list
		R := NULL:

		# Walk through the intervals
		x2 := -infinity:
		for i from 1 to nops(pl) by 3 do
			# Get the interval boundaries
			x1 := x2:
			if i < nops(pl) then
				x2 := pl[i+1]:
			else
				x2 := infinity:
			end if:

			# Get the sub-pl regions
			R2 := PwfPlToRegions(pl[i]):

			for j from 1 to nops(R2) do
				R := R, [x1,x2,op(R2[j])]:
			end do:

			# Have a point to our right?
			if i < nops(pl) then
				R2 := PwfPlToRegions(pl[i+2]):

				for j from 1 to nops(R2) do
					R := R, [x2,x2,op(R2[j])]:
				end do:
			end if:
		end do:

		return [R]:
	end proc:


	# Determines interdependencies of variables
	# in a region.  Returns a list of sets.  If
	# variable vl[n] depends on variable vl[m] (m < n), then
	# m is in the set dl[n].
	RegionDeps := proc( r, vl )
		local dl, l, i, j, k:

		dl := [{}$nops(vl)]:
		j := 3:
		for i from 2 to nops(vl) do
			for k from 1 to i-1 do
				if has(r[j],vl[k]) or has(r[j+1],vl[k]) then
					dl[i] := dl[i] union {k}:
					dl[i] := dl[i] union dl[k]:
				end if:
			end do:

			j := j + 2:
		end do:

		return dl:
	end:


	# Pivots a monotone 2D region.  There are 23 distinct types
	# of monotone 2D regions (12 unique up to reflection).  Each
	# pivot may fracture the one monotone (x,y) region into up to
	# 5 (y,x) regions.
	PivotRegion2dMonotone := proc( r, vl )
		local fa, fb, ga, gb, fv, gv, R,
			x, y, a, b, f, g, F, fm, gm, i:

		(a,b,f,g,F) := op(r):
		(x,y) := op(vl):

		# SPECIAL CASE
		# f,g are not dependent on x?  Simply
		# swap the orders.
		if not has(f,x) and not has(g,x) then
			return [[f,g,a,b,F]]:
		end if:

		# SPECIAL CASE
		# If first interval is a point (a=b), simply swap
		# order.
		if AreEqual(a,b) then
			fa := SimplifyExpr(limit(f,x=a,right)):
			ga := SimplifyExpr(limit(g,x=a,right)):
			return [[fa,ga,a,a,F]]:
		end if:

		# SPECIAL CASE
		# If second interval is a point (f=g), simply
		# invert the function
		if f = g then
			fv := InvertMonotoneFunction(f,x,a,b,y):
			fa := SimplifyExpr(limit(f,x=a,right)):
			fb := SimplifyExpr(limit(f,x=b,left)):

			if EvalRel(fa<fb) then
				return [[fa,fb,fv,fv,F]]:
			else
				return [[fb,fa,fv,fv,F]]:
			end if:
		end if:

		# If we get here, then a and b are finite,
		# and one of f or g is a function of x.

		# Get function values at endpoints
		fa := SimplifyExpr(limit(f,x=a,right)):
		fb := SimplifyExpr(limit(f,x=b,left)):
		ga := SimplifyExpr(limit(g,x=a,right)):
		gb := SimplifyExpr(limit(g,x=b,left)):

		# Get f slope
		if AreEqual(fa,fb) then
			fm := 0:
		elif EvalRel(fa<fb) then
			fm := 1:
		else
			fm := -1:
		end if:

		# Get g slope
		if AreEqual(ga,gb) then
			gm := 0:
		elif EvalRel(ga<gb) then
			gm := 1:
		else
			gm := -1:
		end if:

		# Get function inverses if we need them
		if fm <> 0 then
			fv := InvertMonotoneFunction(f,x,a,b,y):
		end if:
		if gm <> 0 then
			gv := InvertMonotoneFunction(g,x,a,b,y):
		end if:

		R := NULL:

		if fm<0 and gm<0 then
			if AreEqual(fa,ga) then
				if AreEqual(fb,gb) then
					R := R, [fb,fa,fv,gv]:
				else
					R := R, [fb,gb,fv,b]:
					R := R, [gb,gb,SimplifyExpr(limit(fv,y=gb)),b]:
					R := R, [gb,ga,fv,gv]:
				end if:
			elif AreEqual(fa,gb) then
				R := R, [ga,gb,fv,b]:
				R := R, [gb,gb,a,b]:
				R := R, [gb,ga,a,gv]:
			elif EvalRel(fa<gb) then
				R := R, [fb,fa,fv,b]:
				R := R, [fa,fa,a,b]:
				R := R, [fa,gb,a,b]:
				R := R, [gb,gb,a,b]:
				R := R, [gb,ga,a,gv]:
			else
				R := R, [fb,gb,fv,b]:
				R := R, [gb,gb,SimplifyExpr(limit(fv,y=gb)),b]:
				R := R, [gb,fa,fv,gv]:
				R := R, [fa,fa,a,SimplifyExpr(limit(gv,y=fa))]:
				R := R, [fa,ga,a,gv]:
			end if:
		elif fm<0 and gm=0 then
			if AreEqual(fa,ga) then
				R := R, [fb,fa,fv,b]:
			else
				R := R, [fb,fa,fv,b]:
				R := R, [fa,fa,a,b]:
				R := R, [fa,ga,a,b]:
			end if:
		elif fm<0 and gm>0 then
			if AreEqual(fa,ga) then
				R := R, [fb,fa,fv,b]:
				R := R, [fa,fa,a,b]:
				R := R, [fa,gb,gv,b]:
			else
				R := R, [fb,fa,fv,b]:
				R := R, [fa,fa,a,b]:
				R := R, [fa,ga,a,b]:
				R := R, [ga,ga,a,b]:
				R := R, [ga,gb,gv,b]:
			end if:
		elif fm=0 and gm<0 then
			if AreEqual(fb,gb) then
				R := R, [gb,ga,a,gv]:
			else
				R := R, [fa,gb,a,b]:
				R := R, [gb,gb,a,b]:
				R := R, [gb,ga,a,gv]:
			end if:
		elif fm=0 and gm=0 then
			R := R, [fa,ga,a,b]:
		elif fm=0 and gm>0 then
			if AreEqual(fa,ga) then
				R := R, [ga,gb,gv,b]:
			else
				R := R, [fa,ga,a,b]:
				R := R, [ga,ga,a,b]:
				R := R, [ga,gb,gv,b]:
			end if:
		elif fm>0 and gm<0 then
			if AreEqual(fb,gb) then
				R := R, [fa,fb,a,fv]:
				R := R, [fb,fb,a,b]:
				R := R, [gb,ga,a,gv]:
			else
				R := R, [fa,fb,a,fv]:
				R := R, [fb,fb,a,b]:
				R := R, [fb,gb,a,b]:
				R := R, [gb,gb,a,b]:
				R := R, [gb,ga,a,gv]:
			end if:
		elif fm>0 and gm=0 then
			if AreEqual(fb,gb) then
				R := R, [fa,fb,a,fv]:
			else
				R := R, [fa,fb,a,fv]:
				R := R, [fb,fb,a,b]:
				R := R, [fb,gb,a,b]:
			end if:
		elif fm>0 and gm>0 then
			if AreEqual(fb,gb) then
				if AreEqual(fa,ga) then
					R := R, [fa,fb,gv,fv]:
				else
					R := R, [fa,ga,a,fv]:
					R := R, [ga,ga,a,SimplifyExpr(limit(fv,y=ga))]:
					R := R, [ga,gb,gv,fv]:
				end if:
			elif AreEqual(fb,ga) then
				R := R, [fa,fb,a,fv]:
				R := R, [fb,fb,a,b]:
				R := R, [ga,gb,gv,b]:
			elif EvalRel(fb<ga) then
				R := R, [fa,fb,a,fv]:
				R := R, [fb,fb,a,b]:
				R := R, [fb,ga,a,b]:
				R := R, [ga,ga,a,b]:
				R := R, [ga,gb,gv,b]:
			else
				R := R, [fa,ga,a,fv]:
				R := R, [ga,ga,a,SimplifyExpr(limit(fv,y=ga))]:
				R := R, [ga,fb,gv,fv]:
				R := R, [fb,fb,SimplifyExpr(limit(gv,y=fb)),b]:
				R := R, [fb,gb,gv,b]:
			end if:
		end if:

		# Tack back on the region function
		R := [R]:
		R := map(i->[op(i),F],R):

		return R:
	end:


	# Pivots a 2D region
	PivotRegion2d := proc( r, vl )
		local R, X, i, j, x, y:

		(x,y) := op(vl):

		# If the first piece is a point, then simply
		# swap the variables!
		if r[1] = r[2] then
			R := [ limit(r[3],vl[1]=r[1]),
				limit(r[4],vl[1]=r[1]),
				r[1],
				r[1],
				SimplifyExpr(subs(x=r[1],r[5])) ]:
			return [R]:
		end if:

		# Find all of the critical points (may fail if there
		# are too many or infinitely many)
		X := {}:
		i := diff(r[3],x)=0:
		if has(i,x) then
			X := X union GetSolutionsInOpenInterval(
				i, vl[1], r[1], r[2] ):
		end if:
		i := diff(r[4],x)=0:
		if has(i,x) then
			X := X union GetSolutionsInOpenInterval(
				i, vl[1], r[1], r[2] ):
		end if:

		# Sort into ascending order
		X := convert( X, list ):
		X := sort( X, (i,j)->EvalRel(i<j) ):
		X := [r[1],op(X),r[2]]:

		# Now, pivot each monotone domain
		R := NULL:
		for i from 1 to nops(X)-1 do
			if i > 1 then
				R := R,[subs(x=X[i],r[3]),subs(x=X[i],r[4]),
					X[i],X[i],r[5]]:
			end if:
			R := R, op(PivotRegion2dMonotone(
				[X[i],X[i+1],op(3..5,r)], vl )):
		end do:
		R := [R]:

		return R:
	end:


	# Pivots a region.  Given a region described in the
	# variable order [x,y,...,z], then redescribe it as the
	# union of a FINITE set of regions in the variable order
	# [z,x,y,...].  Will fail if can not be represented in
	# finite variable order, or if the region is not able to
	# be pivoted.
	PivotRegion := proc( r, vl )
		local R, R2, V, i, j, k, r2, vl2, F:

		# Ensure varlist and region are consistent
		if 2*nops(vl)+1 <> nops(r) then
			error "region/varlist dimension mismatch":
		end if:

		# Determine the interdependencies of the last
		# variable
		V := RegionDeps( r, vl ):
		V := V[nops(V)]:

		# Get an index to the last variables boundaries
		j := 2*nops(vl)-1:

		R := NULL:

		# Trivial case: 0 dependencies, 1 dimensional
		if nops(V) = 0 then
			r2 := [r[j],r[j+1],op(1..j-1,r),r[j+2]]:
			R := R, r2:
		# 1 dependency, 2 dimensional
		elif nops(V) = 1 then
			# Get an index to the dependent variable
			k := 2*V[1]-1:

			# Pivot the appropriate 2D region
			R2 := PivotRegion2d(
				[r[k],r[k+1],r[j],r[j+1],r[j+2]],
				[vl[V[1]],vl[nops(vl)]] ):

			# Upscale these to the full dimension
			for i from 1 to nops(R2) do
				r2 := [op(1..2,R2[i]),op(1..k-1,r),
					op(3..4,R2[i]),op(k+2..j-1,r),r[j+2]]:
				R := R, r2:
			end do:
		end if:

		# Were we able to pivot the region?
		if R = NULL then
			error "unable to pivot region":
		end if:

		R := [R]:

		# Grab the new last variable, and the
		# region function.
		V := vl[nops(vl)-1]:
		F := r[nops(r)]:

		# Ensure that the new last variable has no
		# occurrence of V in it.
		if has(F,V) then
			j := nops(r):
			for i from 1 to nops(R) do
				if R[i][j-2] = R[i][j-1] then
					R[i][j] := SimplifyExpr(subs(V=R[i][j-2],R[i][j])):
				end if:
			end do:
		end if:

		# Return the set of pivoted regions
		return R:
	end:


	# A comparison function used for sorting regions
	# TODO: Make these use EvalRel
	RegionCompare := proc( r1, r2 )
		if AreEqual(r1[1],r2[1]) then
			if EvalRel(r1[2] <= r2[2]) then
				return true:
			else
				return false:
			end if:
		else
			if EvalRel(r1[1] < r2[1]) then
				return true:
			else
				return false:
			end if:
		end if:
	end proc:


	# Converts a list of REGIONs and a VARLIST to
	# a PWFPL.
	RegionsToPwfPl := proc( _R, vl )
		local pwfpl, pl, i1, i2, x, R2, vl2, R, xv, inf:

		# BASE CASE: Zero dim?
		if nops(vl) = 0 then
			#if nops(_R) > 1 then
			#	error "duplicated piece":
			#end if:
			return _R[1][1]:
		end if:

		R := _R:

		x := vl[1]:
		vl2 := subsop(1=NULL,vl):

		# Do any substitutions/evaluations for sorting
		# purposes
		for i1 from 1 to nops(R) do
			i2 := R[i1][1..2]:
			if nargs > 2 then
				# TODO: Use SimplifyExpr here?
				i2 := SimplifyExpr(subs(args[3..nargs],i2)):
			end if:
			R[i1] := [op(i2),op(R[i1])]:
		end do:

		# Sort the region
		R := sort(R,RegionCompare):

		# TODO: Use EvalRel here?
		# Look for contiguous overlapping regions and split them
		i1 := 2:
		while i1 <= nops(R) do
			# Do these successive regions overlap?
			if R[i1-1][1..2] <> R[i1][1..2] and SimpleEvalRel(R[i1][1] < R[i1-1][2]) then
				if R[i1][1] = R[i1-1][1] then
					# a..b
					# a.....c
					# [a,b]:[a,c] -> [a,b]:[a,b],[b,b],[b,c]

					R2 := [op(5..nops(R[i1]),R[i1])]:
					
					R := [op(R), [op(1..4,R[i1-1]),op(R2)]]:
					R := [op(R), [R[i1-1][2]$2,R[i1-1][4]$2,op(R2)]]:
					R := [op(R), [R[i1-1][2],R[i1][2],R[i1-1][4],R[i1][4],op(R2)]]:

					R := subsop(i1=NULL,R):
				elif R[i1][2] = R[i1-1][2] then
					# a.....b
					#    c..b
					# [a,b]:[c,b] -> [a,c],[c,c],[c,b]:[c,b]

					R2 := [op(5..nops(R[i1-1]),R[i1-1])]:
					
					R := [op(R), [R[i1-1][1],R[i1][1],R[i1-1][3],R[i1][3],op(R2)]]:
					R := [op(R), [R[i1][1]$2,R[i1][3]$2,op(R2)]]:
					R := [op(R), [op(1..4,R[i1]),op(R2)]]:

					R := subsop(i1-1=NULL,R):
				elif SimpleEvalRel(R[i1][2] < R[i1-1][2]) then
					# a........b
					#    c..d
					# [a,b]:[c,d] -> [a,c],[c,c],[c,d],[d,d],[d,b]:[c,d]

					R2 := [op(5..nops(R[i1-1]),R[i1-1])]:
					
					R := [op(R), [R[i1-1][1],R[i1][1],R[i1-1][3],R[i1][3],op(R2)]]:
					R := [op(R), [R[i1][1]$2,R[i1][3]$2,op(R2)]]:
					R := [op(R), [op(1..4,R[i1]),op(R2)]]:
					R := [op(R), [R[i1][2]$2,R[i1][4]$2,op(R2)]]:
					R := [op(R), [R[i1][2],R[i1-1][2],R[i1][4],R[i1-1][4],op(R2)]]:

					R := subsop(i1-1=NULL,R):
				else
					# a.....b
					#    c.....d
					# [a,b]:[c,d] -> [a,c],[c,c],[c,b]:[c,b],[b,b],[b,d]

					R2 := [op(5..nops(R[i1-1]),R[i1-1])]:
					
					R := [op(R), [R[i1-1][1],R[i1][1],R[i1-1][3],R[i1][3],op(R2)]]:
					R := [op(R), [R[i1][1]$2,R[i1][3]$2,op(R2)]]:
					R := [op(R), [R[i1][1],R[i1][3],R[i1-1][2],R[i1-1][4],op(R2)]]:

					R2 := [op(5..nops(R[i1]),R[i1])]:

					R := [op(R), [R[i1][1],R[i1][3],R[i1-1][2],R[i1-1][4],op(R2)]]:
					R := [op(R), [R[i1-1][2]$2,R[i1-1][4]$2,op(R2)]]:
					R := [op(R), [R[i1-1][2],R[i1][1],R[i1-1][4],R[i1][3],op(R2)]]:
					
					R := subsop(i1-1=NULL,i1=NULL,R):
				end if:
				R := sort(R,RegionCompare):
				i1 := 1:
			end if:
			i1 := i1 + 1:
		end do:

		# Ensure the regions are contiguous
		for i1 from 2 to nops(R) do
			if R[i1-1][1..2] <> R[i1][1..2] then
				if R[i1-1][2] <> R[i1][1] then
					error "disjoint domains":
				end if:
				if R[i1-1][1] = R[i1-1][2] and R[i1][1] = R[i1][2] then
					error "successive points":
				end if:
				if R[i1-1][1] <> R[i1-1][2] and R[i1][1] <> R[i1][2] then
					print( R, i1 ):

					error "successive intervals":
				end if:
			end if:
		end do:

		# Finally, do the conversion
		pl := NULL:
		i1 := 1:
		while i1 <= nops(R) do
			i2 := i1:
			R2 := NULL:
			while i2 <= nops(R) and R[i1][1..2] = R[i2][1..2] do
				R2 := R2, R[i2][5..nops(R[i2])]:
				i2 := i2 + 1:
			end do:
			R2 := [R2]:

			# Interval?
			if R[i1][1] <> R[i1][2] then
				xv := evalf(GetPointInOpenInterval(R[i1][1],R[i1][2])):
			else
				xv := evalf(R[i1][1]):
			end if:

			# Recurse on the sub-block
			R2 := RegionsToPwfPl(R2,vl2,args[3..nargs],x=xv):

			# Output old style piece
			pl := pl, [R[i1][3],R2,R[i1][4]]:

			i1 := i2:
		end do:
		pl := [pl]:

		# Generate an 'infinity' piece
		inf := infinity:
		for i1 from 1 to nops(vl2) do
			inf := [inf]:
		end do:

		# Need pieces at beginning?
		if pl[1][1] <> -infinity then
			if pl[1][1] <> pl[1][3] then
				pl := [[pl[1][1],inf,pl[1][1]],op(pl)]:
			end if:
			pl := [[-infinity,inf,pl[1][1]],op(pl)]:
		end if:

		# Need pieces at end?
		i1 := nops(pl):
		if pl[i1][3] <> infinity then
			if pl[i1][1] <> pl[i1][3]  then
				pl := [op(pl),[pl[i1][3],inf,pl[i1][3]]]:
			end if:
			pl := [op(pl),[pl[i1][3],inf,infinity]]:
		end if:

		# Convert old style piecelist to PWFPL.
		pwfpl := NULL:
		for i1 from 1 to nops(pl) by 2 do
			pwfpl := pwfpl, pl[i1][2]:
			if i1 < nops(pl) then
				pwfpl := pwfpl, pl[i1+1][1], pl[i1+1][2]:
			end if:
		end do:
		pwfpl := [pwfpl]:

		return pwfpl:
	end proc:


	# Fixes a 1D PWF after a pivot operation
	`PWFPL/PivotFix1D` := proc( _pl, vl )
		local pl, i, pw, x:

		# No need to do any work?
		x := selectfun(_pl,abs):
		x := select(has,x,vl[1]):
		if nops(x) = 0 then
			return _pl:
		end if:

		# Get the PWFPL and main var
		pl := _pl:
		x := vl[1]:

		# Convert this object to a piecewise object
		if nops(pl) = 1 then
			# TODO: Maple sometimes has problems with signum,
			# returning improper simplifications, or being unable
			# to simplify what is obviously simplifiable.
			pw := convert(pl[1],piecewise,x) assuming op(AssumeCur):
		else
			pw := NULL:
			for i from 2 to nops(pl) by 3 do
				pw := pw, x < pl[i], pl[i-1]:
				pw := pw, x = pl[i], pl[i+1]:
			end do:
			pw := pw, pl[i-1]:
			pw := piecewise(pw):
		end if:

		# Convert back to a PWFPL
		pl := PiecewiseToPiecelist(pw,x):

		# Return it!
		return pl:
	end proc:


	# Ensure last dimension PWFs are valid following a pivot
	`PWFPL/PivotFix` := proc( _pl, vl )
		local pl, vl2, i, x, x1, x2:

		# BASE CASE: One dim?
		if nops(vl) = 1 then
			return `PWFPL/PivotFix1D`(_pl,vl):
		end if:

		# Get the piecelist, main var, and sub-varlist
		pl := _pl:
		x := vl[1]:
		vl2 := subsop(1=NULL,vl):

		# Walk through the piecelist
		x2 := -infinity:
		for i from 1 to nops(pl) do
			# Skip points
			if i mod 3 = 2 then
				next:
			end if:

			# Interval?
			if i mod 3 = 1 then
				x1 := x2:
				if i < nops(pl) then
					x2 := pl[i+1]:
				else
					x2 := infinity:
				end if:

				AssumePush({x1<x,x<x2}):

				pl[i] := `PWFPL/PivotFix`(pl[i],vl2):

				AssumePop():
			# Value at a point?
			else
				pl[i] := `PWFPL/PivotFix`(pl[i],vl2):
			end if:
		end do:

		return pl:
	end proc:


	# Determines the values of any boundary points as
	# appropriate for the current variable order.  Gets
	# applied after a pivot operation.
	`PWFPL/FixBoundaryPoints` := proc( _pl, vl )	
		local pl, isbpt, x, vl2, i, inf, n, checklast, b:

		# Get a copy of the piecelist
		pl := _pl:

		# Get optional 'isbpt', defaulting to false.
		if nargs > 2 then
			isbpt := args[3]:
		else
			isbpt := false:
		end if:

		# Get sub-varlist
		x := vl[1]:
		vl2 := subsop(1=NULL,vl):

		# Generate 'infinity'
		inf := infinity:
		for i from 1 to nops(vl2) do
			inf := [inf]:
		end do:

		n := nops(pl):
		checklast := true:

		# Any points at all?
		if n >= 3 then
			# Process first point
			if pl[1] = inf then
				# If we're not sure if this is a boundary
				# point yet, check at this level.
				b := isbpt:
				if not b then
					if pl[4] <> inf then
						b := true:
					end if:
				end if:

				# BASE CASE: One dimensional
				if nops(vl) = 1 then
					# Is this a boundary point? (not a singleton!)
					if b then
						if pl[1] = inf then
							if pl[4] = inf then
								pl[3] := inf:
							else
								pl[3] := limit(pl[4],x=pl[2],right):
							end if:
						else
							pl[3] := limit(pl[1],x=pl[2],left):
						end if:
					end if:
				else
					# Set the sub boundary points
					pl[3] := `PWFPL/FixBoundaryPoints`(pl[3],vl2,b):
				end if:

				# Avoid checking the last point,
				# which is actually the same point.
				if n = 4 then
					checklast := false:
				end if:
			end if:

			# Process last point
			if checklast and pl[n] = inf then
				# If we're not sure if this is a boundary
				# point yet, check at this level.
				b := isbpt:
				if not b then
					if pl[n-3] <> inf then
						b := true:
					end if:
				end if:

				# BASE CASE: One dimensional
				if nops(vl) = 1 then
					# Is this a boundary point? (not a singleton!)
					if b then
						if pl[n-3] = inf then
							if pl[n] = inf then
								pl[n-1] := inf:
							else
								pl[n-1] := limit(pl[n],x=pl[n-2],right):
							end if:
						else
							pl[n-1] := limit(pl[n-3],x=pl[n-2],left):
						end if:
					end if:
				else
					# Set the sub boundary points
					pl[n-1] := `PWFPL/FixBoundaryPoints`(pl[n-1],vl2,b):
				end if:
			end if:
		end if:

		return pl:	
	end proc:


	# Pivots a PWFPL.
	`PWFPL/Pivot` := proc( _pl, _vl )
		local pl, vl, R, R2, i:

		# Avoid work for trivial cases
		if nops(_vl) < 1 then
			return _pl,_vl:
		end if:

		# Get the piecelist
		pl := _pl:
		vl := _vl:

		# Convert the PWF to list of REGIONs.
		R := PwfPlToRegions(pl):

		# Pivot the regions
		R2 := NULL:
		for i from 1 to nops(R) do
			R2 := R2, op(PivotRegion(R[i],vl)):
		end do:
		R2 := [R2]:

		# Pivot the varlist
		vl := [vl[nops(vl)],op(1..nops(vl)-1,vl)]:

		# Convert back to PWFPL
		pl := RegionsToPwfPl(R2,vl):

		# HACK: Sometimes we'll end up with
		# discontinuities on open intervals.
		R := selectfun(pl,{abs,signum,Re,Im,sign,csgn}):
		R := select(has,R,vl[nops(vl)]):
		if nops(R) > 0 then
			pl := `PWFPL/PivotFix`(pl,vl):
		end if:
		
		# HACK: Somestimes we'll have 'boundary points'
		# which have different values depending on the
		# variable order of the limit.  This patches these
		# up to make them consistent with the current variable
		# order.
		pl := `PWFPL/FixBoundaryPoints`(pl,vl):

		return pl,vl:
	end proc:


	#################################################################
	# CONJUGATION CODE                                              #
	#################################################################


	# Recursively does 1D conjugate along last dimension of
	# supplied PWF.
	`PWFPL/ConjRecur` := proc( _pl, vl, y )
		local pl, i, vl2, x, x1, x2:

		# BASE CASE: One dim?
		if nops(vl) = 1 then
			return `PWFPL/Conj1D`(_pl,vl,[y],AssumeCur):
		end if:

		# Get main var and sub-varlist
		x := vl[1]:
		vl2 := subsop(1=NULL,vl):

		# Grab a copy
		pl := _pl:

		# Walk through the piecelist
		x2 := -infinity:
		for i from 1 to nops(pl) do
			# Skip points
			if i mod 3 = 2 then
				next:
			end if:

			# Interval?
			if i mod 3 = 1 then	
				x1 := x2:
				if i < nops(pl) then
					x2 := pl[i+1]:
				else
					x2 := infinity:
				end if:
				AssumePush({x1<x,x<x2}):
			end if:

			# Recursively process sub-pl
			pl[i] := `PWFPL/ConjRecur`(pl[i],vl2,y):

			# Interval?
			if i mod 3 = 1 then
				AssumePop():
			end if:
		end do:

		# Return the conjugated piecelist
		return pl:
	end proc:


	# Negates a given PWFPL
	`PWFPL/Negate` := proc( _pl, vl )
		local pl, vl2, i:

		# BASE CASE: Zero dim?
		if nops(vl) = 0 then
			if has(_pl,infinity) then
				return infinity:
			else
				return -_pl:
			end if:
		end if:

		# Grab a copy
		pl := _pl:

		# Grab the sub-varlist
		vl2 := subsop(1=NULL,vl):

		# Walk through the piecelist
		for i from 1 to nops(pl) do
			# Skip points
			if i mod 3 = 2 then
				next:
			end if:

			# Recursively process the sub-pl
			pl[i] := `PWFPL/Negate`(pl[i],vl2):
		end do:

		# Return the negated PWFPL
		return pl:
	end proc:


	# Conjugates a PWFPL, recursively processing
	# higher dimensional PWFs as necessary.
	`PWFPL/Conj` := proc( _pl, _vl, vl2 )
		local i, j, vl, y, pl:

		# Grab a copy
		pl := _pl:
		vl := _vl:

		# Conjugate along the first dimension
		j := nops(vl):
		y := vl2[j]:
		pl := `PWFPL/ConjRecur`( pl, vl, y ):
		vl[nops(vl)] := y:

		# If more than 1D
		if nops(vl) > 1 then
			# Process each subsequent dimension
			for i from 2 to nops(vl) do
				# Pivot the results of the last round
				(pl,vl) := `PWFPL/Pivot`(pl,vl):

				# Negate the results of the last round
				pl := `PWFPL/Negate`(pl,vl):

				# Get the conjugation variable
				j := j - 1:
				y := vl2[j]:

				# Conjugate along the i'th dimension
				pl := `PWFPL/ConjRecur`(pl,vl,y):
				vl[nops(vl)] := y:
			end do:

			# Do one more pivot (to get back to
			# our original variable order).
			(pl,vl) := `PWFPL/Pivot`(pl,vl):
		end if:

		# Return the fully conjugated function
		return pl:
	end proc:


	# Conjugates (or attempts to, rather) a PWF object.
	Conj := proc( _pwf::PWF, _vl2::{name,VARLIST} )
		options remember:
		local x, pwf, pl, vl, al, vl2:

		# If vl2 is only a name, make it a list
		vl2 := _vl2:
		if type(vl2,name) then
			vl2 := [vl2]:
		end if:
		
		# Grab a copy of the PWF
		pwf := _pwf:

		# Dimensions match?		
		if nops(vl2) <> nops(op(2,pwf)) then
			error "variable count mismatch":
		end if:

		# Ensure the chosen variable does not appear in the PWF
		x := GetVars(pwf):
		x := x intersect convert(vl2,set):
		if nops(x) <> 0 then
			error "choose different variables: %1", x:
		end if:

		# Reset the assumption stack
		AssumeResetStack():
		
		# Break down the PWF
		(pl,vl,al) := op(pwf):

		# Update the assumptions with the new variables
		al := al union map(x->x::real,convert(vl2,set)):

		# Push the assumptions
		AssumePush(al):

		# Do the conjugation
		pl := `PWFPL/Conj`(pl,vl,vl2):

		# Pop the assumption
		AssumePop():

		# Reconstruct the PWF
		pwf := Assume(PWF(pl,vl2,{}),op(al)):

		# Simplify the constructed PWF
		pwf := `PWF/Simplify`(pwf):

		# Return the conjugated PWF
		return pwf:
	end proc:
	protect(Conj):


	#################################################################
	# GENERAL CONVEX ANALYSIS                                       #
	#################################################################


	# Attempts to determine whether or not a given function
	# is convex.  Returns true, false or FAIL.
	Convex := proc( _pwf::PWF )
		options remember:
		local pwf, vl1, vl2:

		try
			# Get the PWF
			pwf := _pwf:

			# Get the varlist
			vl1 := op(2,pwf):

			# Zero dimension?
			if nops(vl1) = 0 then
				error "zero dimension PWF":
			end if:

			# Generate a conjugate varlist
			vl2 := [seq(_Convex||i,i=1..nops(vl1))]:

			# Conjugate twice
			pwf := Conj(pwf,vl2):
			pwf := Conj(pwf,vl1):

			# Test for equality
			return Equal(_pwf,pwf):
		catch :
			# Fail gracefully
			return FAIL:
		end try:
	end proc:
	protect(Convex):


	# Performs infimal convolution on a sequence of PWFs.
	InfConv := proc()
		options remember:
		local pwf, vl1, vl2, pwfl, i:

		# Get the arguments
		pwfl := [args]:

		# Valid arguments?
		if not type(pwfl,list(PWF)) then
			error "expect seq of PWFs":
		end if:

		if nops(pwfl) < 2 then
			error "expect at least 2 arguments":
		end if:

		# TODO: Synchronization

		vl1 := op(2,pwfl[1]):

		# Ensure non-zero dimension
		if nops(vl1) = 0 then
			error "zero dimension PWF":
		end if:

		# Ensure variable orders match
		for i from 2 to nops(pwfl) do
			if vl1 <> op(2,pwfl[i]) then
				error "variable mismatch":
			end if:
		end do:

		# Produce conjugate variables
		vl2 := [seq(_IC||i,i=1..nops(vl1))]:

		# Conjugate the arguments
		for i from 1 to nops(pwfl) do
			pwfl[i] := Conj( pwfl[i], vl2 ):
		end do:

		# Sum the conjugates
		pwf := simplify(`+`(op(pwfl))):

		# Conjugate the response
		pwf := Conj( pwf, vl1 ):

		return pwf:
	end proc:
	protect(InfConv):


	#################################################################
	# MISCELLANEOUS FUNCTIONS                                       #
	#################################################################


	# Calculates the partial conjugate (one-dimensional)
	# along the last dimension.
	PartialConj := proc( pwf::PWF, Y::Or(VARLIST(1),name) )
		local pl, vl, al, y:

		# Extract the variable
		if type(Y,name) then
			y := Y:
		else
			y := Y[1]:
		end if:

		# Ensure y does not occur in the PWF
		if has(pwf,y) then
			error "please choose another variable: %1", y:
		end if:

		# Reset the assumption stack
		AssumeResetStack():

		# Grab the constituents
		(pl,vl,al) := op(pwf):

		# Push the assumptions
		AssumePush(al):

		# Conjugate along the nth dimension
		pl := `PWFPL/ConjRecur`( pl, vl, y ):
		vl[nops(vl)] := y:

		# Pop the assumptions
		AssumePop():

		# Put PWF back together
		pl := Assume(PWF(pl,vl,{}),op(al)):

		# Simplify the resulting PWF
		pl := `PWF/Simplify`(pl):

		# Return the partially conjugated PWD
		return pl:
	end proc:
	protect(PartialConj):


	# Pivots a given PWF.
	Pivot := proc( pwf::PWF )
		local pl, vl, al:

		# Reset the assumption stack
		AssumeResetStack():

		# Grab the constituents
		(pl,vl,al) := op(pwf):

		# Push the assumptions
		AssumePush(al):

		# Pivot
		(pl,vl) := `PWFPL/Pivot`( pl, vl ):

		# Pop the assumptions
		AssumePop():

		# Put PWF back together
		pl := PWF(pl,vl,al):

		# Simplify the resulting PWF
		pl := `PWF/Simplify`(pl):

		# Return the partially conjugated PWD
		return pl:
	end proc:
	protect(Pivot):


	# Converts a one-dimensional PWF object to a corresponding
	# piecewise object.
	unprotect(PwfToPiecewise):
	PwfToPiecewise := proc( pwf::PWF(1) )
		local i, pl, vl, al, pw, x:

		# Extract parts of PWF
		pl,vl,al := op(pwf):

		# SPECIAL CASE: Single definition for whole
		# real line
		if nops(pl) = 1 then
			return pl[1]:
		end if:

		# Get the main variable
		x := vl[1]:

		# Construct the piecewise parameters
		pw := NULL:
		for i from 1 to nops(pl)-3 by 3 do
			pw := pw, x < pl[i+1], pl[i], x = pl[i+1], pl[i+2]:
		end do:
		pw := pw, pl[i]:

		# Create and simplify the piecewise object
		pw := simplify(piecewise(pw)) assuming op(al):

		# Return it!
		return pw:
	end proc:
	protect(PwfToPiecewise):


	# Converts a PWF object to it's corresponding region representation.
	unprotect(PwfToRegions):
	PwfToRegions := proc( pwf::PWF )
		local R, pl, vl, al:

		# Get the pwf parts
		pl,vl,al := op(pwf):

		# Reset the assumption stack
		AssumeResetStack():

		# Push our assumptions
		AssumePush(al):

		# Get the regions
		R := PwfPlToRegions(pl):

		# Pop our assumptions
		AssumePop(al):

		# Return the PWF
		return R:
	end proc:
	protect(PwfToRegions):


	# Converts a PWF object to it's corresponding region representation.
	unprotect(RegionsToPwf):
	RegionsToPwf := proc(
		R::listlist,
		X::Or(VARLIST,name),
		A::set(ASSUMPTION)
		)
		local pwf, pl, vl,al:

		# Get the varlist
		if not type(X,list) then
			vl := [X]:
		else
			vl := X:
		end if:

		# Get assumptions, if any
		if nargs > 2 then
			al := A:
		else
			al := {}:
		end if:

		# Reset the assumption stack
		AssumeResetStack():

		# Push our assumptions
		AssumePush(al):

		# Get the piecelist
		pl := RegionsToPwfPl( R, vl ):

		# Pop our assumptions
		AssumePop(al):

		# Make and return a new PWF
		return Assume(PWF(pl,vl,{}),op(al)):
	end proc:
	protect(RegionsToPwf):


	#################################################################
	# MODULE BODY                                                   #
	#################################################################


	# Start up this module (export typedefs, pretty printers,
	# simplifiers, etc)
	ModuleStartup():

end module:
